﻿///////////////////////////////////////////////////////////////////////////////
// file entities.js
// Copyright (c) 2012 Frédéric J. Rézeau. All rights reserved.
///////////////////////////////////////////////////////////////////////////////

(function (namespace, undefined) {
    "use strict";

    // Base Entity.
    namespace.Entity = function (x, y, radius) {
        if (!(this instanceof namespace.Entity)) {
            throw new Error("Constructor called as a function.");
        }

        this.x = x;                                             // Position x.
        this.y = y;                                             // Position y.
        this.radius = radius;                                   // Bounding circle radius.
        this.vx = 0;                                            // Velocity x.
        this.vy = 0;                                            // Velocity y.
        this.scaleFactor = 1;                                   // Scale factor.
        this.speed = 0;                                         // Current speed.
        this.baseSpeed = 0;                                     // Entity base speed.
        this.rotationSpeed = 0;                                 // Current rotation speed.
        this.angle = 0;                                         // Current angle.
        this.mass = 1;                                          // Entity mass.
        this.gravity = 9.78;                                    // Gravity factor.
        this.direction = namespace.DirectionEnum.None;          // Current direction.
        this.classType = namespace.EntityClassTypeEnum.None;    // Entity type.
        this.state = namespace.EntityStateEnum.Idle;            // Entity state.
        this.isDead = false;                                    // Determine whether the entity is dead.

        // Animation heartbeats.
        this.heartBeats = [
            { "time": 0, "direction": 0, "maxTime": 1.5, "speed1": 2.2, "speed2": 3.3 },                    // Breathing.
            { "time": 0, "direction": 0, "maxTime": 2 + Math.random() * 2, "speed1": 1, "speed2": 2 },      // Alternate face.
            { "time": 0, "direction": 0, "maxTime": 1, "speed1": 1.5, "speed2": 2.5 },                      // Walk1.
            { "time": 0, "direction": 0, "maxTime": 1, "speed1": 3, "speed2": 5 },                          // Walk2.
            { "time": 0, "direction": 0, "maxTime": 1, "speed1": 1, "speed2": 2 },                          // Walk3.
            { "time": 0, "direction": 0, "maxTime": 1, "speed1": 3, "speed2": 6 },                          // Hammer.
            { "time": 0, "direction": 0, "maxTime": 1, "speed1": 6, "speed2": 12 }                          // Hammer 2X.
        ];
    };

    // Update the Entity time-based properties.
    namespace.Entity.prototype.update = function (elapsed) {
        // Update the animation heart beats.
        var i, len, heartbeat;
        for (i = 0, len = this.heartBeats.length; i < len; i+=1) {
            heartbeat = this.heartBeats[i];
            if (heartbeat.direction === 0 && heartbeat.time <= heartbeat.maxTime) {
                heartbeat.time = Math.min(heartbeat.time + elapsed * heartbeat.speed1, heartbeat.maxTime);
                if (heartbeat.time === heartbeat.maxTime) {
                    heartbeat.direction = 1;
                }
            }
            else if (heartbeat.direction === 1 && heartbeat.time >= 0) {
                heartbeat.time = Math.max(heartbeat.time - elapsed * heartbeat.speed2, 0);
                if (heartbeat.time === 0) {
                    heartbeat.direction = 0;
                }
            }
        }
    };

    // Zombie entity.
    namespace.ZombieEntity = function (x, y, radius) {
        if (!(this instanceof namespace.ZombieEntity)) {
            throw new Error("Constructor called as a function.");
        }

        namespace.Entity.call(this, x, y, radius);

        // Set the class type.
        this.classType = namespace.EntityClassTypeEnum.Zombie;

        // Set the speed.
        this.baseSpeed = 0.55;
        this.speed = this.baseSpeed;

        // Randomize the animation times.
        this.heartBeats[2].time = Math.random() * 0.3;
        this.heartBeats[3].time = this.heartBeats[2].time;
        this.heartBeats[2].direction = (Math.random() > 0.5) ? 0 : 1;
        this.heartBeats[3].direction = this.heartBeats[2].direction;

        this.moveCooldown = 0;                                  // Zombie cannot move while on cooldown.
        this.bouncing = false;                                  // Determine whether the zombie is bouncing.
        this.hitPoints = 22;                                    // Current number of hit points.
        this.maxHitPoints = 22;                                 // Maximum number of hit points.
        this.armorCoefficient = 1;                              // Armor coefficient.
        this.canApplyDamage = false;                            // Whether the entity can apply its base damage.
        this.damageFactor = 1;                                  // Damage factor.
        this.headStyle = namespace.HeadStyleEnum.None;          // Head style.
        this.woundedCooldown = 0;                               // Wounded cooldown.
        this.drawHitPoints = true;                              // Draw the hit points bar.
        this.offScreenDirection = namespace.DirectionEnum.None; // Is the zombie entity off screen.
        this.bonus = false;                                     // Bonus entity.
        this.rpgDamageCooldown = 0;
    };

    // Setup the ZombieEntity prototype chain.
    namespace.ZombieEntity.prototype = Object.create(namespace.Entity.prototype);

    // Update the ZombieEntity time-based properties.
    namespace.ZombieEntity.prototype.update = function (elapsed) {
        // Update the wounded cooldown.
        if (this.woundedCooldown > 0) {
            this.woundedCooldown -= elapsed;
            if (this.woundedCooldown <= 0) {
                this.woundedCooldown = 0;
            }
        }

        if (this.rpgDamageCooldown > 0) {
            this.rpgDamageCooldown -= elapsed;
            if (this.rpgDamageCooldown <= 0) {
                this.rpgDamageCooldown = 0;
            }
        }

        // Update the move cooldown.
        if (this.moveCooldown > 0) {
            this.moveCooldown -= elapsed;
            if (this.moveCooldown < 0) {
                this.moveCooldown = 0;
            }
        }

        // Update velocity on x based on current speed.
        if (!this.bouncing && this.moveCooldown === 0 && this.speed !== 0) {
            switch (this.direction) {
                case namespace.DirectionEnum.Left:
                    this.vx = -this.speed;
                    this.state = namespace.EntityStateEnum.Walking;
                    break;
                case namespace.DirectionEnum.Right:
                    this.vx = this.speed;
                    this.state = namespace.EntityStateEnum.Walking;
                    break;
                case namespace.DirectionEnum.None:
                    this.state = namespace.EntityStateEnum.Idle;
                    this.vx = 0;
                    break;
            }
        }
        else if (this.moveCooldown > 0) {
            this.state = namespace.EntityStateEnum.Idle;
        }

        // Apply friction on X velocity.
        if (this.bouncing) {
            this.moveCooldown = 0.7 + Math.random() * 0.5;

            if (this.vx > 0) {
                this.vx -= 7 * elapsed;
                if (this.vx <= 0) {
                    this.vx = 0;
                }
            }
            else if (this.vx < 0) {
                this.vx += 7 * elapsed;
                if (this.vx > 0) {
                    this.vx = 0;
                }
            }

            if (this.vx === 0 && this.vy === 0) {
                this.bouncing = false;
            }
        }

        // Apply friction on rotation speed.
        if (this.rotationSpeed > 0) {
            this.rotationSpeed -= 10 * elapsed;
            if (this.rotationSpeed < 0) {
                this.rotationSpeed = 0;
            }
        }
        else if (this.rotationSpeed < 0) {
            this.rotationSpeed += 10 * elapsed;
            if (this.rotationSpeed > 0) {
                this.rotationSpeed = 0;
            }
        }

        // Restore the entity angle.
        if (this.rotationSpeed === 0 && this.angle !== 0) {
            if (this.angle > 0) {
                this.angle -= (Math.PI * 2) * elapsed;
                if (this.angle < 0) {
                    this.angle = 0;
                }
            }
            else {
                this.angle += (Math.PI * 2) * elapsed;
                if (this.angle > 0) {
                    this.angle = 0;
                }
            }
        }

        // Apply rotation to angle.
        this.angle = this.rotationSpeed ? (this.angle + this.rotationSpeed * (Math.PI * 2) * elapsed) % (Math.PI * 2) : this.angle;

        // Apply gravity.
        this.vy += this.gravity * elapsed;

        // Apply velocity.
        if (this.moveCooldown === 0 || this.bouncing) {
            this.y += this.vy * elapsed;
            this.x += this.vx * elapsed;
        }

        // Update dead flag.
        if (this.hitPoints <= 0) {
            this.isDead = true;
        }

        namespace.Entity.prototype.update.call(this, elapsed);
    };

    // Human entity.
    namespace.HumanEntity = function (x, y, radius) {
        if (!(this instanceof namespace.HumanEntity)) {
            throw new Error("Constructor called as a function.");
        }

        namespace.Entity.call(this, x, y, radius);

        this.targetScaleFactor = 1;     // Target scale factor for smooth transitions.

        // Weapon related properties.
        this.weapons = [];              // Collection of weapons indexed by category.
        this.reloadCooldown = 0;        // Entity is reloading its weapon.
        this.firingAngle = 0;           // Current firing angle.
        this.firingCooldown = 0;        // Firing cooldown.
        this.firingRecoilTime = 0;      // Simulate weapon recoil.
        this.firingTimeleftOver = 0;    // Preserve damage accuracy on lower frame rate.
        this.weaponOffset = { "x": 0, "y": 0, "tx": 0, "ty": 0 }; // Purely cosmetic to adjust the weapon position.
        this.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;  // Equiped weapon category.

        // Build related properties.
        this.buildDirection = namespace.DirectionEnum.None;
        this.buildData = { "cooldown": 0, "stackObjectType": namespace.StackObjectTypeEnum.None };

        // Associated objects stack.
        this.stack = {};
    };

    // Setup the HumanEntity prototype chain.
    namespace.HumanEntity.prototype = Object.create(namespace.Entity.prototype);

    // Update the HumanEntity time-based properties.
    namespace.HumanEntity.prototype.update = function (elapsed) {
        // Update the current scale factor.
        if (this.targetScaleFactor !== this.scaleFactor) {
            if (this.scaleFactor < this.targetScaleFactor) {
                this.scaleFactor += 3 * elapsed;
                if (this.scaleFactor > this.targetScaleFactor) {
                    this.scaleFactor = this.targetScaleFactor;
                }
            }
            else if (this.scaleFactor > this.targetScaleFactor) {
                this.scaleFactor -= 3 * elapsed;
                if (this.scaleFactor < this.targetScaleFactor) {
                    this.scaleFactor = this.targetScaleFactor;
                }
            }
        }

        // Update the firing cooldown.
        if (this.firingCooldown > 0) {
            this.firingCooldown -= elapsed;
            if (this.firingCooldown <= 0) {
                // Keep track of the left over time here since firing cooldown can have very high rate (e.g. Automatic weapons)
                // we want to preserve the time accuracy at any frame rate.
                this.firingTimeleftOver = this.firingCooldown;
                this.firingCooldown = 0;
            }
        }

        // Firing recoil time for animations.
        if (this.firingRecoilTime > 0) {
            this.firingRecoilTime -= elapsed;
            if (this.firingRecoilTime < 0) {
                this.firingRecoilTime = 0;
            }
        }

        // Reload cooldown time.
        if (this.reloadCooldown > 0) {
            this.reloadCooldown -= elapsed;
            if (this.reloadCooldown < 0) {
                this.reloadCooldown = 0;
            }
        }

        // Apply rotation to angle.
        this.angle = 0;

        // Apply gravity.
        this.vy += this.gravity * elapsed;

        // Apply friction.
        if (this.vx > 0) {
            this.vx -= 2 * elapsed;
            if (this.vx <= 0) {
                this.vx = 0;
            }
        }
        else if (this.vx < 0) {
            this.vx += 2 * elapsed;
            if (this.vx > 0) {
                this.vx = 0;
            }
        }

        // Apply velocity.
        this.y += this.vy * elapsed;
        this.x += this.vx * elapsed;

        namespace.Entity.prototype.update.call(this, elapsed);
    };

})(window.ZCJGame = window.ZCJGame || {});
