﻿///////////////////////////////////////////////////////////////////////////////
// file game.js
// Copyright (c) 2012 Frédéric J. Rézeau. All rights reserved.
///////////////////////////////////////////////////////////////////////////////

(function (namespace, undefined) {
    "use strict";

    // Resolution dependent variables.
    var resolutionX,
        resolutionY,
        scaleUnit,
        assetsSubFolder;

    var canDisplayAd = false;
    var interstitial;

    // Saved data.
    var savedData =
        {
            "playerName": "player",
            "levelProgress": [],
            "killCountHighScore": 0,
            "daysCountHighScore": 0,
            "achievements": [],
            "startCount": 1
        };

    var showReviewMsg = false;
    var showThanksMsgTime = 0;

    var maxStoryModeLevel = 25;

    var noAchievements = false;

    // Game variables.
    var optimalRatioX,      // Optimal ratio on X.
        optimalRatioY,      // Optimal ratio on Y.
        fitToScreen,        // Fit the content to screen (stretch regardless of original ratio).
        canvas,             // Canvas.
        context,            // 2D context.

        gamePaused,         // Game is paused.
        gameWasPaused,      // Game was paused.
        currentGameState,   // Current game state.
        lastUpdateTime,     // Last update time for the game loop.
        levelManager,       // Level manager.

        unitCamera,         // Unit scaling camera applying the meter per pixel ratio.
        primaryCamera,      // The primary camera holds the translate transformation matrix.
        secondaryCamera,    // The secondary camera holds the rotation and scaling transformation matrix.
        hudCamera,          // Stand-alone camera for hud elements.
        cameraTargetZoom,   // Camera target zoom level.
        cameraTargetAngle,  // Camera target angle.
        cameraTarget,       // Current camera target position.
        cameraShakingTime,  // Camera shaking time.
        cameraShakingDir,   // Camera shaking Direction.
        cameraRotationSpeed,// Camera rotation speed.
        slowMotionTime,     // Scene is rendered in slow motion.
        slowMotionSpeed,    // Slow motion speed.
        wantSlowMotion,     // Want slow motion after shaking time.

        ai,                 // AI.

        isPointerDown,      // Pointer is down.
        wasTrial,           // Cache trial mode so we can detect any runtime updates.

        gameLogoTime,      // Game logo time.
        gameIntroTime,      // Game intro time.
        gameMaxIntroTime,   // Game max intro time.
        gameMenuSetupTime,  // Game menu setup time.
        gameMaxMenuSetupTime,  // Game max menu setup time.

        introEndSound = false,

        // Menu related properties.
        currentMenuType,
        menuEntities,
        menuPageIndex,
        menuPageTime,
        menuPageDirection,
        levelItemData,
        menuAchievementItems,
        menuSurvivalItems,
        confirmCommand,
        showCredits,
        itemLockedDialogTime = 0,
        achievementDialogTime = 0,
        achievementDialogMaxTime = 3.2,
        achievementId = -1,
        menuHudItems,
        levelHudItems;

    var gameLoadingData = { "current": 0, "total": 0 };

    var bannerShown = false;
    var adLevel = 0;

    var showReviveScreen = false;

    var desktop = false;

    // Scene animation heartbeats.
    var heartBeats = [
        { "time": 0, "direction": 0, "maxTime": 0.2, "speed1": 2, "speed2": 2 },    // Camera shakes.
        { "time": 0, "direction": 0, "maxTime": 0.5, "speed1": 1, "speed2": 1.5 },  // Title heartbeat.
        { "time": 0, "direction": 0, "maxTime": 0.7, "speed1": 1, "speed2": 2 },     // Weapon heartbeat.
        { "time": 0, "direction": 0, "maxTime": 0.7, "speed1": 2, "speed2": 4 }     // Car heartbeat.
    ];

    namespace.getScaleUnit = function () { return scaleUnit; };
    namespace.getCurrentGameState = function () { return currentGameState; };
    namespace.getCanvasWidth = function () { return canvas.width; };
    namespace.getCanvasHeight = function () { return canvas.height; };
    namespace.getLogoTime = function () { return gameLogoTime; };
    namespace.getIntroTime = function () { return gameIntroTime; };
    namespace.getMaxIntroTime = function () { return gameMaxIntroTime; };
    namespace.getMenuSetupTime = function () { return gameMenuSetupTime; };
    namespace.getMaxMenuSetupTime = function () { return gameMaxMenuSetupTime; };
    namespace.getTitleHeartbeat = function () { return heartBeats[1]; };
    namespace.getWeaponHeartbeat = function () { return heartBeats[2]; };
    namespace.getCarHeartbeat = function () { return heartBeats[3]; };
    namespace.getMenuEntities = function () { return menuEntities; };
    namespace.getCurrentMenuType = function () { return currentMenuType; };
    namespace.getLevelItemData = function () { return levelItemData; };
    namespace.getMenuAchievementItems = function () { return menuAchievementItems; };
    namespace.getMenuSurvivalItems = function () { return menuSurvivalItems; };
    namespace.getMenuPageIndex = function () { return menuPageIndex; };
    namespace.getMenuPageTime = function () { return menuPageTime; };
    namespace.getMenuPageDirection = function () { return menuPageDirection; };
    namespace.getShowCredits = function () { return showCredits; };
    namespace.setShowCredits = function (value) { showCredits = value; };
    namespace.getConfirmCommand = function () { return confirmCommand; };
    namespace.getLockedDialogTime = function () { return itemLockedDialogTime; };
    namespace.getMenuHudItems = function () { return menuHudItems; };
    namespace.getLevelHudItems = function () { return levelHudItems; };
    namespace.getSavedData = function () { return savedData; };
    namespace.getCurrentLevel = function () { return (levelManager ? levelManager.currentLevel : 0); };
    namespace.setPaused = function (value) { gamePaused = value; };
    namespace.getAchievementDialogTime = function () { return achievementDialogTime; };
    namespace.getAchievementDialogMaxTime = function () { return achievementDialogMaxTime; };
    namespace.getAchievementId = function () { return achievementId; };
    namespace.getGameLoadingData = function () { return gameLoadingData; };
    namespace.getShowReviewMsg = function () { return showReviewMsg; };
    namespace.getShowThanksMsgTime = function () { return showThanksMsgTime; };
    namespace.getShowReviveScreen = function () { return showReviveScreen; };

    namespace.getShowBoosters = function () { return (levelManager && levelManager.showBoosters) ? true : false; };
    namespace.getShowProducts = function () { return (levelManager && levelManager.showProducts) ? true : false; };
    namespace.setShowBoosters = function (val) { if (levelManager) { levelManager.showBoosters = val; } };
    namespace.setShowProducts = function (val) { if (levelManager) { levelManager.showProducts = val; } };

    namespace.loadData = function (data) {
        if (data === undefined) {
            if (typeof window.external !== "undefined" && typeof window.external.notify !== "undefined") {
                window.external.notify("commandLoadData");
            }
            else {
                // Uncomment to reset the game data.
                //window.localStorage.setItem("gameData", "{ \"playerName\": \"player\", \"levelProgress\": [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1], \"killCountHighScore\" : 0, \"daysCountHighScore\" : 0, \"achievements\": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0] }");
                //window.localStorage.setItem("gameData", "{ \"playerName\": \"player\", \"levelProgress\": [2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3], \"killCountHighScore\" : 0, \"daysCountHighScore\" : 0, \"achievements\": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0] }");

                var itemData = window.localStorage.getItem("gameData");
                if (itemData === null || itemData === "") {
                    itemData = "{ \"playerName\": \"player\", \"levelProgress\": [2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3], \"killCountHighScore\" : 0, \"daysCountHighScore\" : 0, \"achievements\": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], \"startCount\": 1 }";
                }
                namespace.loadData(itemData);
            }
        }
        else {
            savedData = JSON.parse(data);

            if (savedData.gold === undefined) {
                savedData.gold = 100;
            }

            if (savedData.lives === undefined) {
                savedData.lives = 3;
            }

            if (savedData.refillTime === undefined) {
                savedData.refillTime = 0;
            }

            if (!savedData.startCount) {
                savedData.startCount = 1;
            }
            else {
                if (savedData.levelProgress[2] === namespace.LevelStatusEnum.Passed && savedData.startCount < 3) {
                    savedData.startCount = 3;
                }
                else {
                    savedData.startCount++;
                }
            }

            // Update achievements.
            var count = 0;
            for (var i = 0; i < savedData.achievements.length; i += 1) {
                if (i === 0 && savedData.levelProgress[15] === namespace.LevelStatusEnum.Passed) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 4 && savedData.levelProgress[19] === namespace.LevelStatusEnum.Passed) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 11 && savedData.killCountHighScore >= 100) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 12 && savedData.killCountHighScore >= 200) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 5 && savedData.daysCountHighScore >= 1) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 6 && savedData.daysCountHighScore >= 3) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
                else if (i === 7 && savedData.levelProgress[1] === namespace.LevelStatusEnum.Passed) {
                    savedData.achievements[i] = namespace.LevelStatusEnum.Passed;
                }
            }
        }
    };

    namespace.saveData = function () {
        if (typeof window.external !== "undefined" && typeof window.external.notify !== "undefined") {
            window.external.notify("commandSaveData:" + JSON.stringify(savedData));
        }
        else {
            window.localStorage.setItem("gameData", JSON.stringify(savedData));
        }
    };

    namespace.playAchievementToaster = function (id) {
        if (achievementDialogTime === 0) {
            achievementDialogTime = achievementDialogMaxTime;
            achievementId = id;
            namespace.SoundManager.playAward();
        }
    };

    var onLoadIntroImage = function () {
        if (currentGameState !== namespace.GameStateEnum.GameLogo) {
            namespace.resetDisplay();
            context.drawImage(introImage, 0, 0, resolutionX, resolutionY);
        }
    };

    // Initialize the game.
    var assetsRes = namespace.AssetResolutionEnum.LoRes; // 854
    namespace.initialize = function () {

        // Determine the assets resolution dynamically.
        if (window.innerWidth < 683) {
            assetsRes = namespace.AssetResolutionEnum.VVLoRes; // 480
        }
        else if (window.innerWidth < 854) {
            assetsRes = namespace.AssetResolutionEnum.VLoRes; // 683
        }

        switch (assetsRes) {
            case namespace.AssetResolutionEnum.HiRes:
                resolutionX = 1366;         // Base resolution on X.
                resolutionY = 768;          // Base resolution on Y.
                scaleUnit = 1 / 128;        // Meter to pixel ratio.
                assetsSubFolder = "1366";
                break;
            case namespace.AssetResolutionEnum.LoRes:
                resolutionX = 854;
                resolutionY = 480;
                scaleUnit = 1 / 80;
                assetsSubFolder = "854";
                break;
            case namespace.AssetResolutionEnum.VLoRes:
                resolutionX = 683;
                resolutionY = 384;
                scaleUnit = 1 / 64;
                assetsSubFolder = "683";
                break;
            case namespace.AssetResolutionEnum.VVLoRes:
                resolutionX = 480;
                resolutionY = 270;
                scaleUnit = 1 / 45;
                assetsSubFolder = "480";
                break;
        }

        // Load the data.
        namespace.loadData();

        fitToScreen = false;

        confirmCommand = namespace.CommandEnum.None;

        // Pre-compute the offsets for tiles rendering.
        namespace.setupTileOffsets();

        // Initialize the canvas and retrieve the 2d context.
        canvas = document.createElement("canvas");
        canvas.id = "canvas";
        //canvas.style.cssText = "idtkscale:ScaleAspectFit;";  // CocoonJS extension

        document.body.appendChild(canvas);
        context = canvas.getContext("2d");

        // Bind the canvas events.
        if (!desktop) {
            canvas.addEventListener("touchstart", pointerDown, false);
        }
        else {
            canvas.addEventListener("mousedown", pointerDown, false);
        }

        // Create the cameras.
        unitCamera = new namespace.Camera();
        primaryCamera = new namespace.Camera();
        secondaryCamera = new namespace.Camera();
        hudCamera = new namespace.Camera();

        // Setup the hud camera.
        hudCamera.sx = 1;
        hudCamera.sy = 1;
        hudCamera.update();

        namespace.resetDisplay();

        // Load the loading screen resources.
        namespace.Resources.loadingScreenImage = new Image();
        namespace.Resources.loadingScreenImage.onload = onLoadingScreenReady;
        namespace.Resources.loadingScreenImage.src = "assets/img/" + assetsSubFolder + "/loadingscreen.png";

        currentGameState = namespace.GameStateEnum.GamePreLoading;

        // Start the game loop.
        runGame();
    };

    function onLoadStep(currentStep, totalSteps) {
        gameLoadingData = { "current": currentStep, "total": totalSteps };
    }

    // Reset the game display.
    namespace.resetDisplay = function () {
        var scaleX, scaleY;

        // Reset the canvas buffer dimensions to resolution.
        canvas.width = resolutionX;
        canvas.height = resolutionY;

        if (fitToScreen) {
            optimalRatioX = window.innerWidth / canvas.width;
            optimalRatioY = window.innerHeight / canvas.height;
        }
        else {
            // Compute the optimal screen ratio.
            scaleX = window.innerWidth / canvas.width;
            scaleY = window.innerHeight / canvas.height;
            optimalRatioX = optimalRatioY = Math.min(scaleX, scaleY);
        }

        // Adjust the canvas size.
        canvas.style.width = resolutionX * optimalRatioX + "px";
        canvas.style.height = resolutionY * optimalRatioY + "px";

        // Center the canvas on screen.
        //$("canvas").center();
        centerElement(canvas);

        // Reset the unit camera.
        if (unitCamera) {
            unitCamera.sx = 1 / scaleUnit;
            unitCamera.sy = 1 / scaleUnit;
            unitCamera.update();
        }
    };

    function centerElement(elem) {
        elem.style.position = "absolute";
        elem.style.left = Math.max(0, ((window.innerWidth - resolutionX * optimalRatioX) / 2)) + "px";
        elem.style.top = Math.max(0, ((window.innerHeight - resolutionY * optimalRatioY) / 2)) + "px";
    }

    // Apply the HUD cameras to the context.
    namespace.applyHudCameras = function () {
        context.setTransform.apply(context, unitCamera.matrix);
        context.transform.apply(context, hudCamera.matrix);
    };

    // Apply the scene cameras to the context.
    namespace.applySceneCameras = function () {
        context.setTransform.apply(context, unitCamera.matrix);
        context.transform.apply(context, primaryCamera.matrix);
        context.transform.apply(context, secondaryCamera.matrix);
    };

    // Apply the scene near parallax cameras.
    namespace.applyVeryFarParallaxCameras = function () {
        context.translate(-secondaryCamera.x * 0.15, -secondaryCamera.y * 0.15);
    };

    // Apply the scene far parallax cameras.
    namespace.applyFarParallaxCameras = function () {
        context.translate(-secondaryCamera.x * 0.3, -secondaryCamera.y * 0.3);
    };

    // Apply the scene near parallax cameras.
    namespace.applyNearParallaxCameras = function () {
        context.translate(-secondaryCamera.x * 0.5, -secondaryCamera.y * 0.5);
    };

    // Apply the scene near parallax cameras.
    namespace.applyVeryNearParallaxCameras = function () {
        context.translate(-secondaryCamera.x * 0.75, -secondaryCamera.y * 0.75);
    };

    // Apply the scene 'here' parallax cameras.
    namespace.applyHereParallaxCameras = function () {
        context.translate(-secondaryCamera.x, -secondaryCamera.y);
    };

    // Project the point from screen coordinates to HUD coordinates.
    namespace.screenToHud = function (point) {
        var camera = new namespace.Camera();
        camera.multiply(unitCamera);
        camera.multiply(hudCamera);
        return camera.screenToCamera(point.x, point.y);
    };

    // Project the point from screen coordinates to scene coordinates.
    namespace.screenToScene = function (point) {
        var camera = new namespace.Camera();
        camera.multiply(unitCamera);
        camera.multiply(primaryCamera);
        camera.multiply(secondaryCamera);
        camera.translate(-secondaryCamera.x, -secondaryCamera.y);
        return camera.screenToCamera(point.x, point.y);
    };

    namespace.cancelConfirmDialog = function () {
        gamePaused = gameWasPaused;
        confirmCommand = namespace.CommandEnum.None;
        for (var u = 0; u < levelHudItems.length;) {
            if (levelHudItems[u].type === namespace.CommandEnum.ConfirmNo || levelHudItems[u].type === namespace.CommandEnum.ConfirmYes) {
                levelHudItems.splice(u, 1);
            }
            else {
                u += 1;
            }
        }
    };

    namespace.displayMenuConfirmDialog = function () {
        if (confirmCommand === namespace.CommandEnum.None) {
            confirmCommand = namespace.CommandEnum.Menu;
            createConfirmDialog();
            gameWasPaused = gamePaused;
            gamePaused = true;
        }
        else {
            namespace.cancelConfirmDialog();
        }
    };

    namespace.showMainScreen = function () {
        if (currentGameState !== namespace.GameStateEnum.GameIntro && currentGameState !== namespace.GameStateEnum.GameLogo && (currentGameState !== namespace.GameStateEnum.GameMenu || (currentGameState === namespace.GameStateEnum.GameMenu && currentMenuType !== namespace.MenuTypeEnum.ShowMain && currentMenuType !== namespace.MenuTypeEnum.None)) && currentGameState !== namespace.GameStateEnum.GameLoading) {
            if (currentGameState !== namespace.GameStateEnum.GameMenu) {
                namespace.SoundManager.playMainMusic();

                bannerShown = false;
            }
            currentGameState = namespace.GameStateEnum.GameMenu;
            menuPageIndex = 0;
            gameMaxMenuSetupTime = 3;
            gameMenuSetupTime = gameMaxMenuSetupTime;
            currentMenuType = namespace.MenuTypeEnum.ShowMain;
            menuPageTime = 0;
            menuPageDirection = namespace.DirectionEnum.None;
        }
    };

    namespace.addGold = function (count) {
        savedData.gold += count;
        namespace.saveData();
        namespace.SoundManager.playAward();

        if (levelManager && levelManager.showProducts) {
            levelManager.showProducts = false;
        }
    };

    function onLoadingScreenReady() {
        var manifest = [
            { "type": "img", "id": "menuScreen", "src": "assets/img/" + assetsSubFolder + "/menuscreen.png" },
            { "type": "img", "id": "hud1Screen", "src": "assets/img/" + assetsSubFolder + "/hud1.png" },
            { "type": "img", "id": "hud2Screen", "src": "assets/img/" + assetsSubFolder + "/hud2.png" },
            { "type": "img", "id": "hud3Screen", "src": "assets/img/" + assetsSubFolder + "/hud3.png" },
            { "type": "img", "id": "loadingScreen", "src": "assets/img/" + assetsSubFolder + "/loadingscreen.png" },
            { "type": "img", "id": "creditsScreen", "src": "assets/img/" + assetsSubFolder + "/credits.png" },
            { "type": "img", "id": "tooltips1", "src": "assets/img/" + assetsSubFolder + "/tooltips01.png" },
            { "type": "img", "id": "tooltips5", "src": "assets/img/" + assetsSubFolder + "/tooltips05.png" },
            { "type": "img", "id": "entities", "src": "assets/img/" + assetsSubFolder + "/entities01.png" },
            { "type": "img", "id": "background1", "src": "assets/img/" + assetsSubFolder + "/background01.png" },
            { "type": "img", "id": "background2", "src": "assets/img/" + assetsSubFolder + "/background02.png" },
            { "type": "img", "id": "objects", "src": "assets/img/" + assetsSubFolder + "/objects01.png" },
            { "type": "img", "id": "effects", "src": "assets/img/" + assetsSubFolder + "/effects01.png" },
            { "type": "img", "id": "backdrop1", "src": "assets/img/" + assetsSubFolder + "/backdrop01.png" },
            { "type": "img", "id": "backdrop2", "src": "assets/img/" + assetsSubFolder + "/backdrop02.png" },
            { "type": "img", "id": "tooltips2", "src": "assets/img/" + assetsSubFolder + "/tooltips02.png" },
            { "type": "img", "id": "tooltips3", "src": "assets/img/" + assetsSubFolder + "/tooltips03.png" },
            { "type": "img", "id": "tooltips4", "src": "assets/img/" + assetsSubFolder + "/tooltips04.png" },
            { "type": "img", "id": "boosters", "src": "assets/img/" + assetsSubFolder + "/boosters.png" },
            { "type": "img", "id": "products", "src": "assets/img/" + assetsSubFolder + "/products.png" },
            { "type": "img", "id": "iap", "src": "assets/img/" + assetsSubFolder + "/iap.png" },
            { "type": "img", "id": "backdrop3", "src": "assets/img/" + assetsSubFolder + "/backdrop03.png" }
        ];

        // Finally, preload the game assets.
        namespace.loadResources(manifest, prepareGame, onLoadStep);
    }

    function prepareGame(images) {

        // Retrieve the preloaded assets.
        namespace.Resources.menuScreenImage = images.menuScreen;
        namespace.Resources.loadingScreenImage = images.loadingScreen;
        namespace.Resources.hud1ScreenImage = images.hud1Screen;
        namespace.Resources.hud2ScreenImage = images.hud2Screen;
        namespace.Resources.hud3ScreenImage = images.hud3Screen;
        namespace.Resources.creditsScreenImage = images.creditsScreen;
        namespace.Resources.entitiesImage = images.entities;
        namespace.Resources.tooltips1Image = images.tooltips1;
        namespace.Resources.tooltips2Image = images.tooltips2;
        namespace.Resources.tooltips3Image = images.tooltips3;
        namespace.Resources.tooltips4Image = images.tooltips4;
        namespace.Resources.tooltips5Image = images.tooltips5;
        namespace.Resources.backgroundImage1 = images.background1;
        namespace.Resources.backgroundImage2 = images.background2;
        namespace.Resources.backdrop01Image = images.backdrop1;
        namespace.Resources.backdrop02Image = images.backdrop2;
        namespace.Resources.backdrop03Image = images.backdrop3;
        namespace.Resources.objectsImage = images.objects;
        namespace.Resources.effectsImage = images.effects;
        namespace.Resources.boostersScreenImage = images.boosters;
        namespace.Resources.productsScreenImage = images.products;
        namespace.Resources.iapImage = images.iap;

        // Load the sounds.
        namespace.SoundManager.initialize(startGame);
    }

    function startGame() {
        var i;

        showCredits = false;

        confirmCommand = namespace.CommandEnum.None;

        // Initialize the menu items.
        menuPageIndex = 0;
        menuPageTime = 0;
        menuPageDirection = namespace.DirectionEnum.None;
        currentMenuType = namespace.MenuTypeEnum.None;
        menuHudItems = [
                { "type": namespace.MenuTypeEnum.Panel, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ShowStoryMode, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ShowSurvivalMode, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ShowOptions, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ShowMain, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.SelectNextPage, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.SelectPrevPage, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ShowCredits, "x": 0, "y": 0, "width": 0, "height": 0 },
                { "type": namespace.MenuTypeEnum.ToggleSound, "x": 0, "y": 0, "width": 0, "height": 0 }
        ];

        // Initialize the level item data.
        levelItemData = [];
        for (i = 0; i < savedData.levelProgress.length; i += 1) {
            levelItemData.push({ "level": i + 1, "x": 0, "y": 0, "status": savedData.levelProgress[i] });
        }

        // Initialize the achievement items.
        menuAchievementItems = [];
        for (i = 0; i < 15; i += 1) {
            menuAchievementItems.push({ "x": 0, "y": 0 });
        }

        // Initialize the survival items.
        menuSurvivalItems = [{ "level": 1000, "x": 0, "y": 0, "status": namespace.LevelStatusEnum.Locked }, { "level": 1001, "x": 0, "y": 0, "status": namespace.LevelStatusEnum.Locked }];
        updateSurvivalLevelStatus();

        // Setup the hud camera.
        hudCamera.sx = 1;
        hudCamera.sy = 1;
        hudCamera.update();

        // Play the logo.
        gameLogoTime = 0.1;
        gameMaxIntroTime = 11;
        gameIntroTime = 11;
        currentGameState = namespace.GameStateEnum.GameLogo;

        wasTrial = namespace.isTrialLicense();

        namespace.SoundManager.playMainMusic();

        // Prepare the ads.
        if (typeof Cocoon !== "undefined" && Cocoon.Ad && Cocoon.Ad.MoPub) {
            interstitial = Cocoon.Ad.MoPub.createInterstitial();

            interstitial.on("load", function () {
                console.log("Interstitial loaded");
            });

            interstitial.on("fail", function () {
                console.log("Interstitial failed");
            });

            interstitial.on("show", function () {
                console.log("Interstitial shown");
            });

            interstitial.on("dismiss", function () {
                console.log("Interstitial dismissed");
            });
        }
    }

    function updateLevelStatus(wonLevelId) {
        var item, i, len = levelItemData.length, makeNextAvail = 0;
        for (i = 0; i < len; i += 1) {
            item = levelItemData[i];
            if (wonLevelId === item.level) {
                item.status = namespace.LevelStatusEnum.Passed;
                savedData.levelProgress[i] = item.status;
                makeNextAvail = item.level + 1;
            }
            if (makeNextAvail === item.level && item.status === namespace.LevelStatusEnum.Locked && item.level <= maxStoryModeLevel) {
                item.status = namespace.LevelStatusEnum.Available;
                savedData.levelProgress[i] = item.status;
            }
        }
    }

    function updateSurvivalLevelStatus() {
        for (var i = 0; i < levelItemData.length; i += 1) {
            // Level 6 unlocks first challenge.
            if (levelItemData[i].level === 6 && (levelItemData[i].status === namespace.LevelStatusEnum.Passed)) {
                menuSurvivalItems[0].status = namespace.LevelStatusEnum.Available;
            }
                // Level 10 unlocks second challenge.
            else if (levelItemData[i].level === 13 && (levelItemData[i].status === namespace.LevelStatusEnum.Passed)) {
                menuSurvivalItems[1].status = namespace.LevelStatusEnum.Available;
            }
        }
    }

    function runGame() {
        // Calculate the elapsed time since previous frame.
        var now = new Date().getTime(), elapsed = Math.min((now - (lastUpdateTime || now)) / 1000, 0.05), normalElapsed = elapsed;
        lastUpdateTime = now;

        // Request the next animation frame.
        window.requestAnimationFrame(runGame);

        // Apply slow motion time.
        if (slowMotionTime > 0) {
            elapsed *= slowMotionSpeed;
        }

        // Reset the context to identity matrix and clear.
        context.setTransform(1, 0, 0, 1, 0, 0);

        if (currentGameState === namespace.GameStateEnum.GamePreLoading) {
            namespace.drawLoader(context);
        }
        else {
            if (currentGameState === namespace.GameStateEnum.GameMenu) {
                context.fillStyle = "#ecaa54";

                if (savedData.startCount === 3) {
               //     showReviewMsg = true;
                    savedData.startCount++;
                }
            }
            //context.fillStyle = "black";
            context.clearRect(0, 0, canvas.width, canvas.height);

            if (currentGameState === namespace.GameStateEnum.GameMenu) {
                context.fillRect(0, 0, canvas.width, canvas.height);
            }

            // Update the frame.
            update(elapsed, normalElapsed);

            // Draw the frame.
            namespace.draw(context, levelManager);
        }
    }

    function loadLevel(level) {
        var manifest;

        // Reset the ads.
        adLevel = 0;

        gamePaused = false;
        isPointerDown = false;

        confirmCommand = namespace.CommandEnum.None;

        // Initialize the level manager.
        levelManager = new namespace.levelManager(level);

        // Create the AI.
        ai = new namespace.AI();

        // Display loading screen.
        currentGameState = namespace.GameStateEnum.GameLoading;

        // Preload the level images.
        if (levelManager.currentLevel > 0) {
            loadScene();
        }
    }

    function loadScene() {
        var camera, size;

        // Camera settings.
        cameraShakingTime = 0;
        slowMotionTime = 0;
        slowMotionSpeed = 0.4;
        cameraRotationSpeed = 0.3;
        wantSlowMotion = false;
        cameraTargetAngle = 0;

        if (!levelManager.primaryWeaponAssignmentDone || !levelManager.secondaryWeaponAssignmentDone || levelManager.newItem !== namespace.StackObjectTypeEnum.None) {
            cameraTargetZoom = 0.8;
            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY + 0.5 };
        }
        else {
            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - 0.7 };
            cameraTargetZoom = 0.6;
        }

        camera = new namespace.Camera();
        camera.multiply(unitCamera);
        size = camera.screenToCamera(canvas.width, canvas.height);
        primaryCamera.x = size.x / 2 - cameraTarget.x;
        primaryCamera.y = size.y / 2 - cameraTarget.y;

        secondaryCamera.angle = 0;
        secondaryCamera.sx = cameraTargetZoom;
        secondaryCamera.sy = cameraTargetZoom;

        // If challenge levels are selected, retrieve the current maximum level in story mode
        // and use its hud items setup.
        if (levelManager.currentLevel === 1000 || levelManager.currentLevel === 1001 || (levelManager.currentLevel >= 21 && levelManager.currentLevel <= 25)) {
            var item, i, len = levelItemData.length;
            var level = 6;
            for (i = 0; i < len; i += 1) {
                item = levelItemData[i];
                if (item.status !== namespace.LevelStatusEnum.Passed) {
                    break;
                }
                level = Math.max(level, i + 1);
            }
            createHudItems(level);
        }
        else {
            createHudItems();
        }

        namespace.SoundManager.playLevelMusic();

        // Ready to run.
        currentGameState = namespace.GameStateEnum.GameRun;
    }

    function createHudItems(level) {
        var size;
        if (levelManager) {
            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

            var currentLevel = level;
            if (level === undefined) {
                currentLevel = levelManager.currentLevel;
            }

            if (levelManager.currentLevel > 5 && namespace.isTrialLicense()) {
                levelHudItems = [
                    { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": size.y - 1, "size": 0.95 },
                    { "selected": false, "enabled": true, "type": namespace.CommandEnum.BuyGame, "x": 4.3, "y": 0.8, "size": 1.8 }
                ];
            }
            else {

                switch (currentLevel) {
                    case 1:
                        levelHudItems = [];
                        break;
                    case 2:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        break;
                    case 3:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 4:
                    case 5:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 6:
                    case 7:
                    case 8:
                    case 9:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 10:
                    case 11:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 12:
                    case 13:
                    case 14:
                    case 15:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 16:
                    case 17:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    case 18:
                    case 19:
                    case 20:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addWeaponAssignmentItems(currentLevel, size);
                        break;
                    default:
                        levelHudItems = [
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.RepairCrate, "direction": namespace.DirectionEnum.Left },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.SpikeCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleWoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.DoubleMetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MetalCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.RepairCrate, "direction": namespace.DirectionEnum.Right },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 },
                            { "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x - 2, "y": 0.1, "size": 0.95 }
                        ];
                        addNewObjectOrEntityItems(size);

                        if (levelManager.newItem === namespace.StackObjectTypeEnum.None) {
                            addWeaponAssignmentItems(currentLevel, size);
                        }

                        break;
                }

                if (levelManager.boosters[2].activated) {
                    levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.TNTCrate, "direction": namespace.DirectionEnum.Left });
                    levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.TNTCrate, "direction": namespace.DirectionEnum.Right });
                }

                if (levelManager.boosters[3].activated) {
                    levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MountedCrate, "direction": namespace.DirectionEnum.Left });
                    levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.MountedCrate, "direction": namespace.DirectionEnum.Right });
                }
            }
        }
    }

    function addNewObjectOrEntityItems(size) {
        var centerX, centerY;

        // Add the weapon assignment items.
        centerX = size.x / 2 - 0.5;
        centerY = size.y / 2 - 0.5;
        levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.CrateSelectionDone, "x": centerX, "y": size.y + 1, "tx": centerX, "ty": size.y - 1.1, "size": 1 });
    }

    function addWeaponAssignmentItems(currentLevel, size) {
        var centerX, centerY;

        // Add the weapon assignment items.
        centerX = size.x / 2 - 0.5;
        centerY = size.y / 2 - 0.5;
        levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Ready, "x": centerX, "y": size.y + 1, "tx": centerX, "ty": size.y - 1.1, "size": 1 });
        switch (currentLevel) {
            case 4:
            case 5:
            case 6:
            case 7:
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Left, "locked": true });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Left, "locked": true });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Right, "locked": true });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Right, "locked": true });
                break;
            case 8:
            case 9:
            case 10:
            case 11:
            case 12:
            case 13:
            case 14:
            case 15:
            case 16:
            case 17:
            case 18:
            case 19:
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Left, "locked": true });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Right, "locked": true });
                break;
            case 20:
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Right, "locked": false });
                break;
            default:
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX - 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Left, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.1, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.Rifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Primary, "weaponType": namespace.WeaponTypeEnum.ShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": true, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.4, "ty": centerY - 1.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Uzi, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY - 0.5, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoRifle, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 3, "ty": centerY + 0.6, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.AutoShotGun, "direction": namespace.DirectionEnum.Right, "locked": false });
                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.AssignWeapon, "x": centerX, "y": centerY, "tx": centerX + 2.7, "ty": centerY + 1.7, "size": 1, "weaponCategory": namespace.WeaponCategoryEnum.Secondary, "weaponType": namespace.WeaponTypeEnum.Rpg, "direction": namespace.DirectionEnum.Right, "locked": false });
                break;
        }
    }

    function checkForVictoryCondition() {
        var victory = false;
        if (currentGameState !== namespace.GameStateEnum.GameWon && currentGameState !== namespace.GameStateEnum.GameLost) {
            switch (levelManager.currentLevel) {
                case 1:
                    victory = (levelManager.entities.length === 0 && levelManager.currentLevelStage === namespace.LevelStageEnum.None);
                    break;
                case 2:
                    victory = (ai.phase === 8 && levelManager.entities.length === 0);
                    break;
                case 3:
                case 4:
                case 5:
                case 6:
                case 7:
                case 8:
                case 9:
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                case 17:
                case 18:
                case 19:
                case 20:
                    victory = (ai.phase === 9 && levelManager.entities.length === 0);
                    break;
                case 21:
                    victory = (levelManager.killCount > 49) ? true : false;
                    break;
                case 22:
                    victory = (levelManager.killCount > 99) ? true : false;
                    break;
                case 23:
                    victory = (levelManager.killCount > 149) ? true : false;
                    break;
                case 24:
                    victory = (levelManager.killCount > 199) ? true : false;
                    break;
                case 25:
                    victory = (levelManager.killCount > 249) ? true : false;
                    break;
                case 1000:
                case 1001:
                    victory = false;
                    break;
                default:
                    // High score based. No winning condition.
                    victory = false;
                    break;
            }
        }
        return victory;
    }

    function checkForAchievementCondition() {
        if (noAchievements) {
            return;
        }
        var achievementEarned = false;
        if (slowMotionTime === 0) {
            switch (levelManager.currentLevel) {
                case 2:
                    if (savedData.achievements[7] !== namespace.LevelStatusEnum.Passed) {
                        if (ai.phase === 8) {
                            savedData.achievements[7] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(7);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 3:
                    if (savedData.achievements[1] !== namespace.LevelStatusEnum.Passed) {
                        if (ai.phase === 8 && levelManager.noScratch === true) {
                            savedData.achievements[1] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(1);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 4:
                case 5:
                case 6:
                case 7:
                    break;
                case 8:
                    if (savedData.achievements[3] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon && levelManager.survivalRage === false) {
                            savedData.achievements[3] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(3);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 9:
                    break;
                case 10:
                    if (savedData.achievements[9] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon && levelManager.rightPlayer.stack.objects.length === 3 && levelManager.leftPlayer.stack.objects.length === 3) {
                            savedData.achievements[9] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(9);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 11:
                    break;
                case 12:
                    if (savedData.achievements[10] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon && levelManager.rightPlayer.stack.objects.length === 0 && levelManager.leftPlayer.stack.objects.length === 0) {
                            savedData.achievements[10] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(10);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 13:
                    break;
                case 14:
                    if (savedData.achievements[8] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.metalCount >= 14) {
                            savedData.achievements[8] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(8);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 15:
                    break;
                case 16:
                    if (savedData.achievements[0] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon) {
                            savedData.achievements[0] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(0);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 17:
                    break;
                case 18:
                    if (savedData.achievements[2] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon && levelManager.usedSpiked && levelManager.usedTurret) {
                            savedData.achievements[2] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(2);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 19:
                    if (savedData.achievements[13] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.TNTCount >= 7) {
                            savedData.achievements[13] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(13);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 20:
                    if (savedData.achievements[4] !== namespace.LevelStatusEnum.Passed) {
                        if (currentGameState === namespace.GameStateEnum.GameWon) {
                            savedData.achievements[4] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(4);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 21:
                case 22:
                case 23:
                case 24:
                case 25:
                case 1000:
                    if (savedData.achievements[11] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.killCount >= 100) {
                            savedData.achievements[11] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(11);
                            achievementEarned = true;
                        }
                    }
                    else if (savedData.achievements[12] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.killCount >= 200) {
                            savedData.achievements[12] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(12);
                            achievementEarned = true;
                        }
                    }
                    break;
                case 1001:
                    if (savedData.achievements[5] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.daysCount >= 1) {
                            savedData.achievements[5] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(5);
                            achievementEarned = true;
                        }
                    }
                    else if (savedData.achievements[6] !== namespace.LevelStatusEnum.Passed) {
                        if (levelManager.daysCount >= 3) {
                            savedData.achievements[6] = namespace.LevelStatusEnum.Passed;
                            namespace.playAchievementToaster(6);
                            achievementEarned = true;
                        }
                    }
                    break;
                default:
                    break;
            }
        }

        // To top it all off achievement.
        if (achievementEarned && savedData.achievements[14] !== namespace.LevelStatusEnum.Passed) {
            var count = 0;
            for (var i = 0; i < savedData.achievements.length; i += 1) {
                count += (savedData.achievements[i] === namespace.LevelStatusEnum.Passed) ? 1 : 0;
            }
            if (count >= 10) {
                savedData.achievements[14] = namespace.LevelStatusEnum.Passed;
            }
        }
    };

    function showFullScreenAd(ignoreLevelstatus) {
        if (ignoreLevelstatus && canDisplayAd) {
            canDisplayAd = false;

            if (interstitial) {
                interstitial.show();
            }
        }
        else if (!ignoreLevelstatus && adLevel === 0) {
            adLevel = levelManager.currentLevel;

            if (interstitial) {
                interstitial.show();
            }
        }
    }

    function showBannerAd() {
        if (assetsRes === namespace.AssetResolutionEnum.LoRes) {
            if (!bannerShown) {
                bannerShown = true;
            }
        }
    }

    function update(elapsed, normalElapsed) {
        var i, len, heartbeat;

        if (namespace.isTrialLicense() !== wasTrial) {
            wasTrial = namespace.isTrialLicense();
            namespace.showMainScreen();
        }

        if (showThanksMsgTime > 0) {
            showThanksMsgTime -= elapsed;
            if (showThanksMsgTime < 0) {
                showThanksMsgTime = 0;
            }
        }

        // Item locked dialog time.
        if (itemLockedDialogTime > 0) {
            itemLockedDialogTime -= elapsed;
            if (itemLockedDialogTime < 0) {
                itemLockedDialogTime = 0;
            }
        }

        if (achievementDialogTime > 0) {
            achievementDialogTime -= elapsed;
            if (achievementDialogTime < 0) {
                achievementDialogTime = 0;
            }
        }

        if (!gamePaused) {
            // Update the game animation heart beats.
            for (i = 0, len = heartBeats.length; i < len; i += 1) {
                heartbeat = heartBeats[i];
                if (heartbeat.direction === 0 && heartbeat.time <= heartbeat.maxTime) {
                    heartbeat.time = Math.min(heartbeat.time + elapsed * heartbeat.speed1, heartbeat.maxTime);
                    if (heartbeat.time === heartbeat.maxTime) {
                        heartbeat.direction = 1;
                    }
                }
                else if (heartbeat.direction === 1 && heartbeat.time >= 0) {
                    heartbeat.time = Math.max(heartbeat.time - elapsed * heartbeat.speed2, 0);
                    if (heartbeat.time === 0) {
                        heartbeat.direction = 0;
                    }
                }
            }
        }

        // Update the slow motion time.
        if (slowMotionTime > 0) {
            slowMotionTime -= normalElapsed;
            if (slowMotionTime < 0) {
                slowMotionTime = 0;
            }
        }

        // Update the refill time if needed.
        var timeout = 1140;
        if (savedData.refillTime > 0) {
            var epoch = new Date().getTime() / 1000;
            var remaining = (savedData.refillTime - epoch);
            if (remaining < 0) {
                savedData.refillTime = 0;
                savedData.lives = 3;
            }
        }

        // If the number of lives is less than 3 we start the timer.
        if (savedData.refillTime === 0 && savedData.lives < 3) {
            savedData.refillTime = new Date().getTime() / 1000 + timeout;
            namespace.saveData();
        }

        // Update Hud Items on normal elapsed time.
        updateHudItems(normalElapsed);

        if (currentGameState === namespace.GameStateEnum.GameLogo) {
            updateGameLogo(normalElapsed);
        }
        else if (currentGameState === namespace.GameStateEnum.GameIntro) {
            updateGameIntro(normalElapsed);
        }
        else if (currentGameState === namespace.GameStateEnum.GameMenu) {
            updateGameMenu(normalElapsed);
        }
            // Game running.
        else if (currentGameState === namespace.GameStateEnum.GameRun || currentGameState === namespace.GameStateEnum.GameWon || currentGameState === namespace.GameStateEnum.GameLost) {
            if (!gamePaused) {
                // Update the level manager first.
                levelManager.update(elapsed);

                // Update the sound manager.
                namespace.SoundManager.update(elapsed);

                // Update the AI.
                ai.update(elapsed, levelManager);

                // Update the scene components.
                updateEffects(elapsed);
                updateStacks(elapsed);

                updatePlayer(levelManager.leftPlayer, elapsed);
                updatePlayer(levelManager.rightPlayer, elapsed);

                updateEntities(elapsed);

                // Victory condition.
                if (checkForVictoryCondition()) {
                    namespace.SoundManager.playVictory();

                    // Setup the victory effect.
                    slowMotionTime = 3;
                    currentGameState = namespace.GameStateEnum.GameWon;
                    levelManager.leftPlayer.vy = -3;
                    levelManager.rightPlayer.vy = -3;
                    cameraShakingTime = 0;
                    createDyingEntityEffect(levelManager.leftPlayer.x - 2, levelManager.maxY - 1, namespace.DirectionEnum.Left);
                    createDyingEntityEffect(levelManager.rightPlayer.x + 2, levelManager.maxY - 1, namespace.DirectionEnum.Right);

                    levelHudItems = [];

                    updateLevelStatus(levelManager.currentLevel);
                    updateSurvivalLevelStatus();

                    // Save the data.
                    namespace.saveData();
                }

                if (currentGameState === namespace.GameStateEnum.GameRun) {
                    updateGameRun();
                }
            }

            if (currentGameState === namespace.GameStateEnum.GameWon) {
                updateGameWon();
            }
            else if (currentGameState === namespace.GameStateEnum.GameLost) {
                updateGameLost();
            }

            checkForAchievementCondition();

            // Update the cameras.
            updateCameras(elapsed, normalElapsed, cameraTarget);
        }
    }

    function updateGameLogo(elapsed) {
        if (gameLogoTime > 0) {
            gameLogoTime -= elapsed;

            if (gameLogoTime <= 0) {
                currentGameState = namespace.GameStateEnum.GameIntro;
            }
        }
    }

    function updateGameIntro(elapsed) {
        var i, entity, size, offset = 0, goLeft = false, baseScale = 0.47;
        if (gameIntroTime > 0) {
            gameIntroTime -= elapsed;

            if (gameIntroTime < 3 && introEndSound === false) {
                introEndSound = true;
                // namespace.SoundManager.playCarSound();
            }

            if (gameIntroTime <= 0) {
                gameIntroTime = 0;
                gameMaxMenuSetupTime = 3;
                gameMenuSetupTime = gameMaxMenuSetupTime;
                currentGameState = namespace.GameStateEnum.GameMenu;

                size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

                // Create the menu entities.
                menuEntities = [];
                for (i = 0; i < 3; i += 1) {
                    entity = new namespace.ZombieEntity(size.x / 2 - 4 + offset, size.y / 2 - Math.random(), 0.4);
                    entity.direction = goLeft ? namespace.DirectionEnum.Left : namespace.DirectionEnum.Right;
                    entity.scaleFactor = baseScale;
                    entity.baseSpeed = baseScale / 2;
                    entity.speed = entity.baseSpeed;
                    entity.drawHitPoints = false;
                    switch (i) {
                        case 0:
                            entity.headStyle = namespace.HeadStyleEnum.Police;
                            break;
                        case 1:
                            entity.headStyle = namespace.HeadStyleEnum.Punk;
                            break;
                    }
                    menuEntities.push(entity);

                    offset += 0.5;
                    goLeft = !goLeft;
                    baseScale += 0.21;
                }
            }
        }
    }

    function updateGameMenu(elapsed) {
        var i, len, limitMinX, limitMaxX, limitY, size, entity;
        // Update the game menu setup time.
        if (gameMenuSetupTime > 0) {
            gameMenuSetupTime -= elapsed;
            if (gameMenuSetupTime < 0) {
                gameMenuSetupTime = 0;
            }
        }

        // Update the menu page time.
        if (menuPageTime > 0) {
            menuPageTime -= elapsed;
            if (menuPageTime < 0) {
                menuPageTime = 0;
            }
        }

        size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
        limitMinX = size.x / 2 - 4;
        limitMaxX = size.x / 2 - 2;
        limitY = size.y / 2 + 1.7;

        for (i = 0, len = menuEntities.length; i < len; i += 1) {
            entity = menuEntities[i];
            entity.update(elapsed);

            if (entity.moveCooldown > 0 && entity.moveCooldown < 0.5) {
                entity.moveCooldown = 0;
                entity.direction = (entity.direction === namespace.DirectionEnum.Left) ? namespace.DirectionEnum.Right : namespace.DirectionEnum.Left;
            }

            if (entity.x > limitMaxX) {
                entity.moveCooldown = 1.5;
                entity.x = limitMaxX;
            }
            else if (entity.x < limitMinX) {
                entity.moveCooldown = 1.5;
                entity.x = limitMinX;
            }

            // Make sure the entity is always above ground level.
            if (entity.y >= limitY - entity.radius) {
                entity.y = limitY - entity.radius;
                entity.vy = 0;
            }
        }
    }

    function updateGameWon() {
        var maxStackCount, size, centerX;

        // Reset the player and entities.
        levelManager.entities = [];
        levelManager.leftPlayer.buildData.cooldown = 0;
        levelManager.rightPlayer.buildData.cooldown = 0;
        levelManager.leftPlayer.reloadCooldown = 0;
        levelManager.rightPlayer.reloadCooldown = 0;
        levelManager.leftPlayer.firingCooldown = 0;
        levelManager.rightPlayer.firingCooldown = 0;

        if (slowMotionTime > 0) {
            slowMotionSpeed = 0.07;

            if (levelManager.leftPlayer.stack.objects.length < levelManager.rightPlayer.stack.objects.length) {
                cameraTargetAngle = namespace.toRadians(20);
            }
            else {
                cameraTargetAngle = namespace.toRadians(-20);
            }
            cameraTargetZoom = 1.5;
            cameraRotationSpeed = 0.02;
            maxStackCount = Math.max(levelManager.leftPlayer.stack.objects.length, levelManager.rightPlayer.stack.objects.length);
            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - maxStackCount * 1.7 };
        }
        else {
            if (levelManager.rewardItem.won === false && (levelManager.rewardItem.crate !== namespace.StackObjectTypeEnum.None || levelManager.rewardItem.weapon !== namespace.WeaponTypeEnum.None)) {
                namespace.SoundManager.playAward();
                // Add the weapon assignment items.
                levelManager.rewardItem.won = true;
                size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                centerX = size.x / 2 - 0.5;
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.RewardAccepted, "x": centerX, "y": size.y + 1, "tx": centerX, "ty": size.y - 2, "size": 1 });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x / 2 - 1.5, "y": size.y / 2 + 1.1, "size": 0.95 });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.NextLevel, "x": size.x / 2 + 0.5, "y": size.y / 2 + 1.1, "size": 0.95 });
            }
            else if (levelManager.rewardItem.crate === namespace.StackObjectTypeEnum.None && levelManager.rewardItem.weapon === namespace.WeaponTypeEnum.None) {
                size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x / 2 - 1.5, "y": size.y / 2 - 0.2, "size": 0.95 });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.NextLevel, "x": size.x / 2 + 0.5, "y": size.y / 2 - 0.2, "size": 0.95 });
            }

            cameraTargetAngle = 0;
            cameraTargetZoom = 0.55;
            cameraRotationSpeed = 0.05;
            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - 0.9 };

            if (levelManager.survivalRageTime > 0) {
                if (levelManager.rightPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Secondary) {
                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                }
                if (levelManager.leftPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Secondary) {
                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                }
            }
            else {
                if (levelManager.rightPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Primary) {
                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                }
                if (levelManager.leftPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Primary) {
                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                }
            }
        }
    }

    function updateGameLost() {
        levelManager.leftPlayer.buildData.cooldown = 0;
        levelManager.rightPlayer.buildData.cooldown = 0;
        levelManager.leftPlayer.reloadCooldown = 0;
        levelManager.rightPlayer.reloadCooldown = 0;
        if (levelManager.leftPlayer.isDead) {
            levelManager.leftPlayer.direction = namespace.DirectionEnum.Left;
            levelManager.leftPlayer.firingCooldown = 0;
            cameraTarget = { "x": levelManager.leftPlayer.x, "y": levelManager.maxY - 1 };
        }
        if (levelManager.rightPlayer.isDead) {
            levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
            levelManager.rightPlayer.firingCooldown = 0;
            cameraTarget = { "x": levelManager.rightPlayer.x, "y": levelManager.maxY - 1 };
        }

        if (levelManager.levelEndCooldown === 0) {

            if (levelHudItems.length === 0) {
                // levelManager.entities = [];

                var size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x / 2 - 2, "y": size.y / 2 + 1, "size": 0.95 });
                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Restart, "x": size.x / 2 + 1, "y": size.y / 2 + 1, "size": 0.95 });
            }

            if (levelManager.levelEndPauseCooldown === 0) {
                gamePaused = true;
                showFullScreenAd();
            }
            slowMotionSpeed = 0.5;
            slowMotionTime = 0.2;
            cameraTargetZoom = 1.6;
        }
    }

    function updateGameRun() {
        var size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

        if (levelManager.newItem !== namespace.StackObjectTypeEnum.None) {
            cameraTargetZoom = 0.8;
            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY + 0.5 };
        }
        else if (!levelManager.secondaryWeaponAssignmentDone || !levelManager.primaryWeaponAssignmentDone) {
            // Update camera target during weapon assignment.
            if (levelManager.rightPlayer.targetScaleFactor === 1 && levelManager.leftPlayer.targetScaleFactor === 1) {
                cameraTargetZoom = 0.8;
                cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY + 0.5 };
            }
            else if (levelManager.rightPlayer.targetScaleFactor > levelManager.leftPlayer.targetScaleFactor) {
                cameraTargetZoom = 0.73;
                cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 + 0.7, "y": levelManager.maxY + 0.5 };
            }
            else if (levelManager.rightPlayer.targetScaleFactor < levelManager.leftPlayer.targetScaleFactor) {
                cameraTargetZoom = 0.73;
                cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 - 0.7, "y": levelManager.maxY + 0.5 };
            }
        }
        else {

            if (levelManager.survivalRageTime > 0) {
                if (levelManager.rightPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Secondary) {
                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                    levelManager.rightPlayer.reloadCooldown = 0.3;
                    namespace.SoundManager.playReloadSound();
                }
                if (levelManager.leftPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Secondary) {
                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                    levelManager.leftPlayer.reloadCooldown = 0.3;
                    namespace.SoundManager.playReloadSound();
                }
            }
            else {
                if (levelManager.rightPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Primary) {
                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                    levelManager.rightPlayer.reloadCooldown = 0.3;
                    namespace.SoundManager.playReloadSound();
                }
                if (levelManager.leftPlayer.currentWeaponCategory !== namespace.WeaponCategoryEnum.Primary) {
                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                    levelManager.leftPlayer.reloadCooldown = 0.3;
                    namespace.SoundManager.playReloadSound();
                }
            }

            // Show the add if needed.
            if (levelManager.currentLevel > 1) {
                //showBannerAd();
            }

            // Update camera target when no level stage is specified.
            if (levelManager.currentLevelStage === namespace.LevelStageEnum.None) {
                if (levelManager.levelStartCooldown > 0) {
                    // Game level is on cooldown.
                    cameraTargetZoom = 0.73;
                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY + 1 };
                }
                else if (levelManager.leftPlayer.direction === namespace.DirectionEnum.Right && levelManager.rightPlayer.direction === namespace.DirectionEnum.Right) {
                    // Both players are looking to the right.
                    cameraTargetZoom = 0.8;
                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 + 1.5, "y": levelManager.maxY - 0.9 };
                }
                else if (levelManager.leftPlayer.direction === namespace.DirectionEnum.Left && levelManager.rightPlayer.direction === namespace.DirectionEnum.Left) {
                    // Both players are looking to the left.
                    cameraTargetZoom = 0.8;
                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 - 1.5, "y": levelManager.maxY - 0.9 };
                }
                else {
                    // Players are looking to different directions.
                    cameraTargetZoom = 0.73;
                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - 0.9 };
                }
            }
            else {
                // Deletegate camera target assignment to level and stage.
                switch (levelManager.currentLevel) {
                    case 1:
                        cameraTargetZoom = 0.8;
                        if ((levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialZombiesComing && levelManager.tutorialCooldown < 1.5) || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialReinforceDefenses) {
                            if (levelHudItems.length === 1) {
                                levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 0.5, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left });
                                levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left });
                            }
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTapToContinue) {
                            levelHudItems = [];
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstWave) {
                            levelHudItems = [];
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstWaveDone) {
                            levelHudItems = [];
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialMoreZombiesComing || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWaveDone) {
                            levelHudItems = [];
                        }
                        else if ((levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstKill || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSelectSurvivalRage) && levelManager.tutorialCooldown === 0) {
                            gamePaused = true;
                        }

                        // Add the menu item if needed.
                        var menuItemPresent = false;
                        for (var i = 0; i < levelHudItems.length; ++i) {
                            if (levelHudItems[i].type === namespace.CommandEnum.Menu) {
                                menuItemPresent = true;
                                break;
                            }
                        }

                        if (menuItemPresent === false) {
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.Menu, "x": size.x - 1, "y": 0.1, "size": 0.95 });
                        }

                        if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialStart) {
                            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - 0.9 };
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialMoreZombiesComing && levelManager.tutorialCooldown > 1 && levelManager.tutorialCooldown < 4) {
                            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 + 9, "y": levelManager.maxY - 0.9 };
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialZombiesComing && levelManager.tutorialCooldown > 1.5) {
                            cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 - 9, "y": levelManager.maxY - 0.9 };
                        }
                        else {
                            if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWave || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWaveDone || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialHelpingHand || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstKill || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSelectSurvivalRage || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstWaveFinishJob) {
                                cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 + 1.5, "y": levelManager.maxY - 0.9 };
                            }
                            else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamWork || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialThirdWave || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamCamera || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamDefenses) {
                                if (levelManager.leftPlayer.direction === namespace.DirectionEnum.Right && levelManager.rightPlayer.direction === namespace.DirectionEnum.Right) {
                                    cameraTargetZoom = 0.8;
                                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 + 1.5, "y": levelManager.maxY - 0.9 };
                                }
                                else if (levelManager.leftPlayer.direction === namespace.DirectionEnum.Left && levelManager.rightPlayer.direction === namespace.DirectionEnum.Left) {
                                    cameraTargetZoom = 0.8;
                                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 - 1.5, "y": levelManager.maxY - 0.9 };
                                }
                                else {
                                    cameraTargetZoom = 0.73;
                                    cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2, "y": levelManager.maxY - 0.9 };
                                }
                            }
                            else {
                                cameraTarget = { "x": levelManager.minX + (levelManager.maxX - levelManager.minX) / 2 - 1.5, "y": levelManager.maxY - 0.9 };
                            }
                        }
                        break;
                    default:
                        break;
                }
            }
        }

        if (levelManager.reachedTop) {
            if (levelManager.leftPlayer.stack.objects.length === 1 || levelManager.rightPlayer.stack.objects.length === 1) {
                showReminder();
            }
        }
    }

    function updateEffects(elapsed) {
        var i, j, u, stack, entity, object, effectRemoved, effect;
        var effectsToAdd = [];

        // Update the effects.
        for (i = 0; i < levelManager.effects.length;) {
            effect = levelManager.effects[i];
            effectRemoved = false;

            // Update.
            effect.update(elapsed);

            // Self destroyable effect.
            if (effect.selfDestroyTime === 0) {
                levelManager.effects.splice(i, 1);
                continue;
            }

            // Effect touched the ground.
            if (effect.y > levelManager.maxY) {
                effect.y = levelManager.maxY;
                if (effect.vy > 0.5) {
                    effect.vy = -effect.vy * effect.mass; // Small bounce.
                }
            }

            // Effect is active.
            if (effect.activationCooldown === 0 && (effect.selfDestroyTime > 1 || effect.selfDestroyTime === -1)) {
                // Effect is on the ground.
                if (effect.y === levelManager.maxY) {
                    switch (effect.type) {
                        case namespace.EffectTypeEnum.Warhead:
                        case namespace.EffectTypeEnum.Bullet:
                        case namespace.EffectTypeEnum.MultiBullet:
                            levelManager.effects.splice(i, 1);
                            effectRemoved = true;
                            break;
                    }
                }
                else {
                    switch (effect.type) {
                        case namespace.EffectTypeEnum.Warhead:
                        case namespace.EffectTypeEnum.Bullet:
                        case namespace.EffectTypeEnum.MultiBullet:
                            // Check if a stack object is colliding.
                            for (j = 0; j < levelManager.stacks.length; ++j) {
                                stack = levelManager.stacks[j];
                                for (u = 0; u < stack.objects.length; ++u) {
                                    object = stack.objects[u];
                                    if (object.type !== namespace.StackObjectTypeEnum.TNTCrate && object.type !== namespace.StackObjectTypeEnum.MetalCrate && object.type !== namespace.StackObjectTypeEnum.WoodCrate && object.type !== namespace.StackObjectTypeEnum.MountedCrate && object.type !== namespace.StackObjectTypeEnum.SpikeCrate && stack.direction !== namespace.DirectionEnum.None && namespace.rectIntersect(effect.x - effect.radius, effect.y - effect.radius, effect.x + effect.radius, effect.y + effect.radius, stack.x - object.radius, object.y - object.radius, stack.x + object.radius, object.y + object.radius)) {
                                        if (stack.vx < 1 && stack.vx > -1 && ((stack.direction === namespace.DirectionEnum.Left && stack.x > levelManager.minX) || (stack.direction === namespace.DirectionEnum.Right && stack.x < levelManager.maxX))) {
                                            stack.vx += (effect.vx > 0) ? 1.25 : -1.25;
                                        }

                                        // Create an impact effect.
                                        levelManager.effects.push(new namespace.ImpactEffect(effect.x + ((effect.vx > 0) ? 0.12 : -0.12), effect.y, 0.15));

                                        // Remove the projectile.
                                        levelManager.effects.splice(i, 1);
                                        effectRemoved = true;
                                        break;
                                    }
                                }

                                // Bail out if the effect was removed.
                                if (effectRemoved) {
                                    break;
                                }
                            }
                            break;
                    }

                    if (!effectRemoved) {
                        // Check if an entity is colliding.
                        for (j = 0; j < levelManager.entities.length; ++j) {
                            entity = levelManager.entities[j];
                            if (entity.hitPoints > 0) {
                                if (namespace.rectIntersect(effect.x - effect.radius, effect.y - effect.radius, effect.x + effect.radius, effect.y + effect.radius, entity.x - entity.radius, entity.y - entity.radius, entity.x + entity.radius, entity.y + entity.radius)) {
                                    switch (effect.type) {
                                        case namespace.EffectTypeEnum.Warhead:
                                        case namespace.EffectTypeEnum.Bullet:
                                        case namespace.EffectTypeEnum.MultiBullet:
                                            var crit = (Math.random() < levelManager.criticalChance) ? true : false;

                                            if (namespace.EffectTypeEnum.Warhead === effect.type) {
                                                entity.hitPoints -= 5;
                                                entity.vy -= 1;
                                                entity.vx = (entity.direction === namespace.DirectionEnum.Left) ? 4 + Math.random() : -(4 + Math.random());
                                                entity.bouncing = true;
                                                effect.count++;
                                            }
                                            else if (crit) {
                                                entity.hitPoints -= effect.value * 2;

                                                var effect1 = new namespace.ProjectileEffect(entity.x, entity.y - 1, 0.5, 0, 0, -3, namespace.EffectTypeEnum.Critical, 0);
                                                effect1.selfDestroyTime = 1;
                                                effect1.startFadeTime = 0.7;
                                                effectsToAdd.push(effect1);
                                                namespace.SoundManager.playCritSound();
                                            }
                                            else {
                                                entity.hitPoints -= effect.value;
                                            }

                                            entity.woundedCooldown = 0.2;

                                            if (entity.hitPoints <= 0 && effect.type === namespace.EffectTypeEnum.MultiBullet) {
                                                entity.bonus = true;
                                            }

                                            // Create a blood splatter effect.
                                            levelManager.effects.push(new namespace.BloodSplatterEffect(effect.x + ((effect.vx > 0) ? 0.12 : -0.12), effect.y, 0.15));

                                            if (namespace.EffectTypeEnum.Warhead !== effect.type || effect.count === 2) {
                                                // Remove the projectile.
                                                levelManager.effects.splice(i, 1);
                                                effectRemoved = true;
                                            }
                                            break;
                                    }
                                }
                            }
                        }
                    }
                }
            }

            if (!effectRemoved) {
                i += 1;
            }
        }

        for (i = 0; i < effectsToAdd.length; i += 1) {
            levelManager.effects.push(effectsToAdd[i]);
        }

        // Update the background effects.
        for (i = 0; i < levelManager.backgroundEffects.length; i += 1) {
            effect = levelManager.backgroundEffects[i];

            // Update.
            effect.update(elapsed);

            // Self destroyable effect.
            if (effect.selfDestroyTime === 0) {
                levelManager.backgroundEffects.splice(i, 1);
                continue;
            }
        }
    }

    function updateCameras(elapsed, normalElapsed, target) {
        // Reset the translation.
        var x = primaryCamera.x, y = primaryCamera.y;
        primaryCamera.x = 0;
        primaryCamera.y = 0;
        primaryCamera.update();

        // Position the target at the center of the canvas.
        var camera = new namespace.Camera();
        camera.multiply(unitCamera);
        var size = camera.screenToCamera(canvas.width, canvas.height);
        primaryCamera.tx = size.x / 2 - target.x;
        primaryCamera.ty = size.y / 2 - target.y;

        // Determine the distance between current and target positions.
        var distance = namespace.distance(x, y, primaryCamera.tx, primaryCamera.ty);
        if (Math.abs(distance) > 0.001) {
            // Interpolation factors on x and y.
            var factorX = (primaryCamera.tx - x) / distance;
            var factorY = (primaryCamera.ty - y) / distance;

            // Camera speed is adjusted based on how far it is from the target.
            var speed = 2;
            var cameraSpeedX = elapsed * distance * speed;
            var cameraSpeedY = elapsed * distance * speed;
            x += cameraSpeedX * factorX;
            y += cameraSpeedY * factorY;
        }
        else {
            x = primaryCamera.tx;
            y = primaryCamera.ty;
        }

        // Update the camera shaking time.
        if (cameraShakingTime > 0) {
            cameraShakingTime -= elapsed;
            if (cameraShakingTime <= 0) {
                cameraShakingTime = 0;
                if (wantSlowMotion) {
                    slowMotionTime = 1;
                }
            }
        }

        // Translation on primary camera.
        primaryCamera.x = x;
        primaryCamera.y = y;

        // Rotation and scaling on secondary camera.
        secondaryCamera.x = size.x / 2 - x;
        secondaryCamera.y = size.y / 2 - y;

        // Zoom.
        if (secondaryCamera.sx > cameraTargetZoom) {
            secondaryCamera.sx -= 0.2 * normalElapsed;
            if (secondaryCamera.sx < cameraTargetZoom) {
                secondaryCamera.sx = cameraTargetZoom;
            }
        }
        if (secondaryCamera.sx < cameraTargetZoom) {
            secondaryCamera.sx += 0.2 * normalElapsed;
            if (secondaryCamera.sx > cameraTargetZoom) {
                secondaryCamera.sx = cameraTargetZoom;
            }
        }
        secondaryCamera.sy = secondaryCamera.sx;

        if (cameraShakingTime > 0) {
            if (cameraShakingDir === namespace.DirectionEnum.Right) {
                secondaryCamera.angle = -0.05 + heartBeats[0].time * 0.25;
            }
            else {
                secondaryCamera.angle = 0.05 - heartBeats[0].time * 0.25;
            }
        }

        // Angle.
        if (secondaryCamera.angle > cameraTargetAngle) {
            secondaryCamera.angle -= cameraRotationSpeed * Math.PI * normalElapsed;
            if (secondaryCamera.angle < cameraTargetAngle) {
                secondaryCamera.angle = cameraTargetAngle;
            }
        }
        else if (secondaryCamera.angle < cameraTargetAngle) {
            secondaryCamera.angle += cameraRotationSpeed * Math.PI * normalElapsed;
            if (secondaryCamera.angle > cameraTargetAngle) {
                secondaryCamera.angle = cameraTargetAngle;
            }
        }

        // Apply changes.
        primaryCamera.update();
        secondaryCamera.update();

        hudCamera.update();
    }

    function updateHudItems(elapsed) {
        var distance, factorX, factorY;
        var i, len, offset, size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
        if (currentGameState === namespace.GameStateEnum.GameMenu) {
            // Compute the menu offset.
            offset = { "x": 5, "y": -((currentMenuType === namespace.MenuTypeEnum.None || currentMenuType === namespace.MenuTypeEnum.ShowMain) ? 0.3 : 2.1) };
            if (gameMenuSetupTime > gameMaxMenuSetupTime - 0.5) {
                offset.y -= (0.5 - (gameMaxMenuSetupTime - gameMenuSetupTime)) * 2;
            }

            // Resize the menu items.
            for (i = 0, len = menuHudItems.length; i < len; i += 1) {
                // Reset the item.
                menuHudItems[i].x = 0;
                menuHudItems[i].y = 0;
                menuHudItems[i].width = 0;
                menuHudItems[i].height = 0;

                // Set the show main item.
                if (currentMenuType !== namespace.MenuTypeEnum.None && currentMenuType !== namespace.MenuTypeEnum.ShowMain && menuHudItems[i].type === namespace.MenuTypeEnum.ShowMain) {
                    menuHudItems[i].x = offset.x + 2;
                    menuHudItems[i].y = offset.y + 2.85;
                    menuHudItems[i].width = 2;
                    menuHudItems[i].height = 0.7;
                }
                else if (menuHudItems[i].type === namespace.MenuTypeEnum.ShowCredits) {
                    menuHudItems[i].x = size.x - 1.2;
                    menuHudItems[i].y = 0.2;
                    menuHudItems[i].width = 1;
                    menuHudItems[i].height = 1;
                }
                else if (menuHudItems[i].type === namespace.MenuTypeEnum.ToggleSound) {
                    menuHudItems[i].x = size.x - 1.2;
                    menuHudItems[i].y = 1.2;
                    menuHudItems[i].width = 1;
                    menuHudItems[i].height = 1;
                }
                else if (menuHudItems[i].type === namespace.MenuTypeEnum.Panel) {
                    menuHudItems[i].x = offset.x;
                    menuHudItems[i].y = offset.y;
                }

                switch (currentMenuType) {
                    case namespace.MenuTypeEnum.None:
                    case namespace.MenuTypeEnum.ShowMain:
                        switch (menuHudItems[i].type) {
                            case namespace.MenuTypeEnum.ShowStoryMode:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 0.6;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 0.95;
                                break;
                            case namespace.MenuTypeEnum.ShowSurvivalMode:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 1.85;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 0.95;
                                break;
                            case namespace.MenuTypeEnum.ShowOptions:
                                menuHudItems[i].x = offset.x - 2.6;
                                if (gameMenuSetupTime > gameMaxMenuSetupTime - 0.5) {
                                    menuHudItems[i].y = offset.y + 5 + ((0.5 - (gameMaxMenuSetupTime - gameMenuSetupTime)) * 2) * 3;
                                }
                                else {
                                    menuHudItems[i].y = offset.y + 5;
                                }
                                //menuHudItems[i].y = offset.y + 5;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 1.2;
                                break;
                        }
                        break;
                    case namespace.MenuTypeEnum.ShowStoryMode:
                        switch (menuHudItems[i].type) {
                            case namespace.MenuTypeEnum.ShowStoryMode:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 2.1;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 0.5;
                                break;
                            case namespace.MenuTypeEnum.SelectPrevPage:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                            case namespace.MenuTypeEnum.SelectNextPage:
                                menuHudItems[i].x = offset.x + 3;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                        }
                        break;
                    case namespace.MenuTypeEnum.ShowSurvivalMode:
                        switch (menuHudItems[i].type) {
                            case namespace.MenuTypeEnum.ShowSurvivalMode:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 2.1;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 0.5;
                                break;
                            case namespace.MenuTypeEnum.SelectPrevPage:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                            case namespace.MenuTypeEnum.SelectNextPage:
                                menuHudItems[i].x = offset.x + 3;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                        }
                        break;
                    case namespace.MenuTypeEnum.ShowOptions:
                        switch (menuHudItems[i].type) {
                            case namespace.MenuTypeEnum.ShowOptions:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 2.1;
                                menuHudItems[i].width = 4;
                                menuHudItems[i].height = 0.5;
                                break;
                            case namespace.MenuTypeEnum.SelectPrevPage:
                                menuHudItems[i].x = offset.x;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                            case namespace.MenuTypeEnum.SelectNextPage:
                                menuHudItems[i].x = offset.x + 3;
                                menuHudItems[i].y = offset.y + 3.8;
                                menuHudItems[i].width = 1;
                                menuHudItems[i].height = 0.7;
                                break;
                        }
                        break;
                }
            }
        }
        else if (currentGameState === namespace.GameStateEnum.GameRun || currentGameState === namespace.GameStateEnum.GameWon) {
            // Update the hud item data.
            for (i = 0; i < levelHudItems.length; i += 1) {
                var hudItem = levelHudItems[i];
                if (hudItem.type === namespace.CommandEnum.SelectStackObject || hudItem.type === namespace.CommandEnum.ToggleAdrenaline) {
                    var speed = 5;
                    if (hudItem.type === namespace.CommandEnum.ToggleAdrenaline) {
                        hudItem.tx = (levelManager.survivalRageTime === 0) ? size.x / 2 - 0.5 : 1.5;
                        hudItem.ty = (levelManager.survivalRageTime === 0) ? size.y - 2 : 0.5;
                        speed = 9;
                    }

                    if (hudItem.x !== hudItem.tx || hudItem.y !== hudItem.ty) {
                        distance = namespace.distance(hudItem.x, hudItem.y, hudItem.tx, hudItem.ty);
                        if (distance > 0.5) {
                            factorX = (hudItem.tx - hudItem.x) / distance;
                            factorY = (hudItem.ty - hudItem.y) / distance;
                            hudItem.x += elapsed * speed * factorX;
                            hudItem.y += elapsed * speed * factorY;
                        }
                        else {
                            hudItem.x = hudItem.tx;
                            hudItem.y = hudItem.ty;
                        }
                    }
                }
                else if (hudItem.type === namespace.CommandEnum.AssignWeapon || hudItem.type === namespace.CommandEnum.Ready || hudItem.type === namespace.CommandEnum.CrateSelectionDone || hudItem.type === namespace.CommandEnum.RewardAccepted) {

                    if (hudItem.type === namespace.CommandEnum.AssignWeapon) {
                        if (!levelManager.primaryWeaponAssignmentDone && hudItem.weaponCategory === namespace.WeaponCategoryEnum.Primary) {
                            hudItem.enabled = true;
                        }
                        else if (levelManager.primaryWeaponAssignmentDone && !levelManager.secondaryWeaponAssignmentDone && hudItem.weaponCategory === namespace.WeaponCategoryEnum.Secondary) {
                            hudItem.enabled = true;
                        }
                        else {
                            hudItem.enabled = false;
                        }
                    }
                    else if (hudItem.type === namespace.CommandEnum.CrateSelectionDone) {
                        if (levelManager.newItem === namespace.StackObjectTypeEnum.None) {
                            hudItem.enabled = false;
                        }
                    }
                    else if (hudItem.type === namespace.CommandEnum.RewardAccepted) {
                        if (levelManager.rewardItem.crate === namespace.StackObjectTypeEnum.None && levelManager.rewardItem.weapon === namespace.WeaponTypeEnum.None) {
                            hudItem.enabled = false;
                        }
                    }
                    else {
                        if (levelManager.primaryWeaponAssignmentDone && levelManager.secondaryWeaponAssignmentDone) {
                            hudItem.enabled = false;
                        }
                    }

                    if (hudItem.enabled && (hudItem.x !== hudItem.tx || hudItem.y !== hudItem.ty)) {
                        distance = namespace.distance(hudItem.x, hudItem.y, hudItem.tx, hudItem.ty);
                        if (distance > 0.5) {
                            factorX = (hudItem.tx - hudItem.x) / distance;
                            factorY = (hudItem.ty - hudItem.y) / distance;
                            hudItem.x += elapsed * 9 * factorX;
                            hudItem.y += elapsed * 9 * factorY;
                        }
                        else {
                            hudItem.x = hudItem.tx;
                            hudItem.y = hudItem.ty;
                        }
                    }
                }
            }
        }
    }

    function updateStacks(elapsed) {
        var x;
        // Reset the mounted crate counter.
        levelManager.mountedCrateCount = 0;
        for (var i = 0; i < levelManager.stacks.length;) {
            var stack = levelManager.stacks[i];

            if (stack.objects.length === 0) {
                levelManager.stacks.splice(i, 1);
                continue;
            }

            if (stack.cooldown > 0) {
                stack.cooldown -= elapsed;
                if (stack.cooldown < 0) {
                    stack.cooldown = 0;
                }
            }

            if (stack.vx > 0) {
                stack.vx -= elapsed * 6;
                if (stack.vx < 0) {
                    stack.vx = 0;
                }
            }
            else if (stack.vx < 0) {
                stack.vx += elapsed * 6;
                if (stack.vx > 0) {
                    stack.vx = 0;
                }
            }

            stack.x += stack.vx * elapsed;

            if (stack.vx !== 0) {
                updateStackCollisions(stack);
            }

            if (stack.direction !== namespace.DirectionEnum.None) {
                if ((stack.direction === namespace.DirectionEnum.Right && stack.x < stack.limitX) || (stack.direction === namespace.DirectionEnum.Left && stack.x > stack.limitX)) {
                    stack.x = stack.limitX;
                }
            }

            var y = levelManager.maxY;
            for (var u = 0; u < stack.objects.length;) {
                var object = stack.objects[u];
                // Get rid of destroyed stack objects.
                if (object.strength <= 0) {
                    namespace.SoundManager.playHumanFailed();
                    if (object.type === namespace.StackObjectTypeEnum.TNTCrate) {

                        if (cameraShakingTime === 0 && slowMotionTime === 0) {
                            cameraShakingDir = (Math.random() > 0.5) ? namespace.DirectionEnum.Left : namespace.DirectionEnum.Right;
                            cameraShakingTime = 0.5;
                            wantSlowMotion = true;
                            heartBeats[0].time = 0;
                            heartBeats[0].direction = 0;

                            levelManager.leftPlayer.vy = -3.5;
                            levelManager.rightPlayer.vy = -3.5;
                        }
                        // Any entity within range will die.
                        var reduceHp = false;
                        var entityCount = 0;
                        for (var j = 0; j < levelManager.entities.length; ++j) {
                            var entity = levelManager.entities[j];
                            var distance = namespace.distance(entity.x, entity.y, stack.x, object.y);
                            if (distance < 4) {
                                if (reduceHp) {
                                    entity.hitPoints *= 0.7;
                                }
                                else {
                                    entity.hitPoints = 0;
                                }
                                entityCount++;
                            }

                            if (entityCount === 2) {
                                reduceHp = true;
                            }
                            else if (entityCount === 4) {
                                break;
                            }
                        }

                        namespace.SoundManager.playBlastSound();

                        for (x = 0; x < 2; ++x) {
                            var speed = 0.1 + Math.random() * 0.5;
                            var directedSpeed = ((Math.random() > 0.5) ? speed : -speed);
                            levelManager.effects.push(new namespace.BigSmokeEffect(stack.x + directedSpeed, object.y + directedSpeed, 0.15));
                        }

                        for (x = 0; x < 2; ++x) {
                            levelManager.effects.push(new namespace.ExplosionEffect(stack.x, object.y, 0.15));
                        }
                    }
                    else {
                        for (x = 0; x < 3; ++x) {
                            levelManager.effects.push(new namespace.SmallSmokeEffect(stack.x, object.y, 0.15));
                        }
                    }
                    stack.objects.splice(u, 1);
                    continue;
                }

                // Update.
                object.update(elapsed);

                if (object.type === namespace.StackObjectTypeEnum.MountedCrate) {
                    levelManager.mountedCrateCount++;
                }

                if (object.type === namespace.StackObjectTypeEnum.MountedCrate && object.firingCooldown === 0) {
                    var closestEntity = getClosestEntity(stack.x, object.y, stack.direction);
                    if (closestEntity.entity !== null && closestEntity.distance < 3 && closestEntity.distance > 0) {

                        var offsetY = -0.2;
                        var position = { "x": stack.x + ((stack.direction === namespace.DirectionEnum.Right) ? 0.3 : -0.3), "y": object.y + offsetY };
                        var angle = namespace.vectorAngle(position.x, position.y, closestEntity.entity.x, closestEntity.entity.y);
                        var offset = namespace.vectorRotate(0.3, (stack.direction === namespace.DirectionEnum.Right) ? offsetY : -offsetY, angle);

                        position.x += offset.x;
                        position.y += offset.y;

                        var effect = new namespace.Effect(position.x, position.y, 0.2);
                        effect.gravity = 0;
                        effect.selfDestroyTime = 2;
                        effect.type = namespace.EffectTypeEnum.Bullet;
                        effect.value = 2.5;
                        effect.angle = namespace.vectorAngle(effect.x, effect.y, closestEntity.entity.x, closestEntity.entity.y);
                        object.firingAngle = effect.angle;
                        object.firingRecoilTime = 0.2;

                        var factorX = (closestEntity.entity.x - position.x) / closestEntity.distance;
                        var factorY = (closestEntity.entity.y - position.y) / closestEntity.distance;
                        effect.vx = 9 * factorX;
                        effect.vy = 9 * factorY;

                        levelManager.effects.push(effect);

                        levelManager.effects.push(new namespace.SmallSmokeEffect(position.x, position.y, 0.15));
                        levelManager.effects.push(new namespace.SmallSmokeEffect(position.x - 0.1, position.y, 0.15));
                        levelManager.effects.push(new namespace.SmallSmokeEffect(position.x + 0.1, position.y, 0.15));

                        object.firingCooldown = 0.3;

                        namespace.SoundManager.playGunSound();
                    }
                }

                // Bouncing on the floor.
                if (object.y >= y - object.radius) {
                    object.y = y - object.radius;
                    if (object.vy > 0.2) {
                        object.vy = -object.vy * 0.3;
                    }
                    else {
                        object.vy = 0;
                    }
                }
                y = object.y - 0.36;
                ++u;
            }

            i += 1;
        }
    }

    function updateStackCollisions(entity) {
        // Check collisions with ennemy objects.
        var entity1 = entity;
        for (var u = 0; u < levelManager.stacks.length; ++u) {
            var entity2 = levelManager.stacks[u];
            if (entity1 !== entity2 && entity1.objects.length > 0 && entity2.objects.length > 0 && entity1.cooldown < 0.9 && entity2.cooldown < 0.8 && entity1.direction !== namespace.DirectionEnum.None && entity2.direction !== namespace.DirectionEnum.None) {

                var position1 = { "x": entity1.x, "y": entity1.objects[0].y };
                var position2 = { "x": entity2.x, "y": entity2.objects[0].y };

                // Check if there is a collision.
                // Multiplications are less computationally expensive than square roots so here we
                // square the sum of the radii instead for comparison.
                var centerDiff = { "x": position1.x - position2.x, "y": position1.y - position2.y };
                var dotProduct = namespace.vectorDotProduct(centerDiff.x, centerDiff.y, centerDiff.x, centerDiff.y);
                var radiusSum = 0.2 + 0.2;
                var radiusSum2 = radiusSum * radiusSum;
                if (dotProduct > radiusSum2) {
                    continue; // No collision.
                }

                var velocity1 = { "x": entity1.vx, "y": entity1.objects[0].vy };
                var velocity2 = { "x": entity2.vx, "y": entity2.objects[0].vy };

                var mass1 = 1;
                var mass2 = 1;

                // Compute the normal of collision - i.e. vector along which momentum is exchanged.
                var distance = Math.sqrt(dotProduct);
                var collisionNormal = namespace.vectorDivide(centerDiff.x, centerDiff.y, distance);

                // Depth of collision divided by 2.
                var collisionDepth = (radiusSum - distance) * 0.5;

                // Push the entities apart so they only touch.
                var positionOffset = namespace.vectorMultiply(collisionNormal.x, collisionNormal.y, collisionDepth);
                position1 = namespace.vectorAdd(position1.x, position1.y, positionOffset.x, positionOffset.y);
                position2 = namespace.vectorSubstract(position2.x, position2.y, positionOffset.x, positionOffset.y);

                entity1.x = position1.x;
                entity2.x = position2.x;

                // Recompute the normal since we moved the entities.
                collisionNormal = namespace.vectorSubstract(position1.x, position1.y, position2.x, position2.y);
                collisionNormal = namespace.vectorNormalize(collisionNormal.x, collisionNormal.y);

                // Find the length of the component of each of the movement vectors along normal.
                // a1 = v1 . N
                // a2 = v2 . N
                var length1 = namespace.vectorDotProduct(velocity1.x, velocity1.y, collisionNormal.x, collisionNormal.y);
                var length2 = namespace.vectorDotProduct(velocity2.x, velocity2.y, collisionNormal.x, collisionNormal.y);
                var optimizedP = (2.0 * (length1 - length2)) / (mass1 + mass2);

                // Calculate new velocity.
                // v1" = v1 - optimizedP * m2 * normal
                var scale1 = namespace.vectorMultiply(collisionNormal.x, collisionNormal.y, mass2 * optimizedP);
                velocity1 = namespace.vectorSubstract(velocity1.x, velocity1.y, scale1.x, scale1.y);
                entity1.vx = velocity1.x;
                entity1.objects[0].vy = velocity1.y;

                // Calculate new velocity.
                // v2" = v2 - optimizedP * m1 * normal
                var scale2 = namespace.vectorMultiply(collisionNormal.x, collisionNormal.y, mass1 * optimizedP);
                velocity2 = namespace.vectorAdd(velocity2.x, velocity2.y, scale2.x, scale2.y);
                entity2.vx = velocity2.x;
                entity2.objects[0].vy = velocity2.y;
            }
        }
    }

    function updatePlayer(player, elapsed) {
        // Update.
        player.update(elapsed);

        if (player.classType === namespace.EntityClassTypeEnum.HumanFemale) {
            switch (player.weapons[player.currentWeaponCategory]) {
                case namespace.WeaponTypeEnum.Rifle:
                    player.weaponOffset.x = 0.5;
                    player.weaponOffset.y = -0.4;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = 0.05;
                    break;
                case namespace.WeaponTypeEnum.ShotGun:
                    player.weaponOffset.x = 0.35;
                    player.weaponOffset.y = -0.4;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = 0.05;
                    break;
                case namespace.WeaponTypeEnum.AutoRifle:
                    player.weaponOffset.x = 0.5;
                    player.weaponOffset.y = -0.4;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = 0.05;
                    break;
                case namespace.WeaponTypeEnum.AutoShotGun:
                    player.weaponOffset.x = 0.5;
                    player.weaponOffset.y = -0.4;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = 0.05;
                    break;
                case namespace.WeaponTypeEnum.Uzi:
                    player.weaponOffset.x = 0.5;
                    player.weaponOffset.y = -0.4;
                    player.weaponOffset.tx = 0.6;
                    player.weaponOffset.ty = -0.02;
                    break;
                case namespace.WeaponTypeEnum.Rpg:
                    player.weaponOffset.x = -0.1;
                    player.weaponOffset.y = -0.5;
                    player.weaponOffset.tx = 0.6;
                    player.weaponOffset.ty = -0.02;
                    break;
            }

            //  player.weaponOffset.y += ((player.direction === namespace.DirectionEnum.Right || player.firingAngle === 0) ? player.firingAngle : Math.PI - player.firingAngle) * 0.2;
        }
        else {
            switch (player.weapons[player.currentWeaponCategory]) {
                case namespace.WeaponTypeEnum.Rifle:
                    player.weaponOffset.x = 0.4;
                    player.weaponOffset.y = 0.3;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = -0.18;
                    break;
                case namespace.WeaponTypeEnum.ShotGun:
                    player.weaponOffset.x = 0.4;
                    player.weaponOffset.y = 0.35;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = -0.18;
                    break;
                case namespace.WeaponTypeEnum.AutoRifle:
                    player.weaponOffset.x = 0.4;
                    player.weaponOffset.y = 0.35;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = -0.25;
                    break;
                case namespace.WeaponTypeEnum.AutoShotGun:
                    player.weaponOffset.x = 0.4;
                    player.weaponOffset.y = 0.35;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = -0.25;
                    break;
                case namespace.WeaponTypeEnum.Uzi:
                    player.weaponOffset.x = 0.4;
                    player.weaponOffset.y = 0.45;
                    player.weaponOffset.tx = 0.8;
                    player.weaponOffset.ty = -0.25;
                    break;
                case namespace.WeaponTypeEnum.Rpg:
                    player.weaponOffset.x = -0.1;
                    player.weaponOffset.y = 0.3;
                    player.weaponOffset.tx = 0.6;
                    player.weaponOffset.ty = -0.02;
                    break;
            }

            //   player.weaponOffset.y += ((player.direction === namespace.DirectionEnum.Right || player.firingAngle === 0) ? player.firingAngle : Math.PI - player.firingAngle) * 0.5;
        }

        // Check that the entity its ground.
        var y = levelManager.maxY;
        if (!player.isDead) {
            if (player.stack.objects.length > 0) {
                y = player.stack.objects[player.stack.objects.length - 1].y - 0.5;
            }
        }

        if (player.y >= y - player.radius && player.vy > 0) {
            player.y = y - player.radius;
            player.vy = 0;
        }

        var buildData = player.buildData;
        if (buildData.cooldown > 0) {
            buildData.cooldown -= elapsed * ((levelManager.survivalRageTime > 0) ? 2 : 1);
            if (levelManager.boosters[1].activated) {
                buildData.cooldown -= elapsed;
            }

            // Cancel the repair if stack length has fallen to 0 or the bottom stack object is at full strength.
            if (buildData.stackObjectType === namespace.StackObjectTypeEnum.RepairCrate && (player.stack.objects.length === 0 || player.stack.objects[0].strength === levelManager.getStackObjectStrength(player.stack.objects[0].type))) {
                buildData.cooldown = 0;
            }

            // Cooldown as elapsed. Create the objects.
            if (buildData.cooldown <= 0) {
                if (buildData.stackObjectType === namespace.StackObjectTypeEnum.RepairCrate) {
                    if (player.stack.objects.length > 0) {
                        player.stack.objects[0].strength = levelManager.getStackObjectStrength(player.stack.objects[0].type);
                    }
                }
                else {
                    addStackObject(buildData.stackObjectType, player);
                }
                buildData.cooldown = 0;
                buildData.stackObjectType = namespace.StackObjectTypeEnum.None;
            }
            else {
                if (player.heartBeats[3].time === 0) {
                    var smokeEffect = new namespace.SmallSmokeEffect(player.x, player.y, 0.15);
                    smokeEffect.selfDestroyTime = 0.6;
                    smokeEffect.startFadeTime = 0.7;
                    levelManager.effects.push(smokeEffect);
                    // levelManager.effects.push(new namespace.ImpactEffect(player.x + 0.3, player.y, 0.15));
                    levelManager.effects.push(new namespace.ImpactEffect(player.x - 0.3, player.y, 0.15));
                    levelManager.effects.push(new namespace.ImpactEffect(player.x + 0.3, player.y + 0.5, 0.15));

                    smokeEffect = new namespace.SmallSmokeEffect(player.x + 0.5, player.y + 0.2, 0.15);
                    smokeEffect.startFadeTime = 0.9;
                    smokeEffect.selfDestroyTime = 0.5;
                    levelManager.backgroundEffects.push(smokeEffect);

                    smokeEffect = new namespace.SmallSmokeEffect(player.x + 0.2, player.y + 0.6, 0.15);
                    smokeEffect.startFadeTime = 0.9;
                    smokeEffect.selfDestroyTime = 0.9;
                    levelManager.effects.push(smokeEffect);

                    namespace.SoundManager.playHammerSound();
                }

                player.reloadCooldown = 0.3;
            }
        }

        var closestEntity = null;
        var canFire = false;
        if (currentGameState !== namespace.GameStateEnum.GameLost && player.reloadCooldown === 0 && buildData.cooldown === 0 && player.firingCooldown === 0) {
            closestEntity = getClosestEntity(player.x, player.y, player.direction);
            if (closestEntity.entity !== null && closestEntity.distance < 10 && closestEntity.distance > 0) {
                canFire = true;
            }
        }

        if (player.firingRecoilTime > 0 || player.firingCooldown > 0) {
            player.weaponOffset.y += player.stack.objects.length * 0.05;
        }

        if (canFire && closestEntity) {
            player.firingRecoilTime = 0.3;

            var position = { "x": player.x + ((player.direction === namespace.DirectionEnum.Right) ? player.weaponOffset.x : -player.weaponOffset.x), "y": player.y + player.weaponOffset.y };
            var angle = namespace.vectorAngle(position.x, position.y, closestEntity.entity.x, closestEntity.entity.y);
            var offset = namespace.vectorRotate(player.weaponOffset.tx, (player.direction === namespace.DirectionEnum.Right) ? player.weaponOffset.ty : -player.weaponOffset.ty, angle);

            position.x += offset.x;
            position.y += offset.y;

            // Interpolation factors.
            var factorX = (closestEntity.entity.x - position.x) / closestEntity.distance;
            var factorY = (closestEntity.entity.y - position.y) / closestEntity.distance;

            var effect = null;
            var firingCooldown = 1;
            var damage = 0;
            var damageMultiplier = (levelManager.boosters[0].activated ? 1.5 : 1);
            switch (player.weapons[player.currentWeaponCategory]) {
                case namespace.WeaponTypeEnum.Rifle:
                    firingCooldown = 0.7;
                    damage = 2;
                    damage *= damageMultiplier;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.Bullet, damage));
                    levelManager.effects.push(new namespace.SmallSmokeEffect(position.x, position.y, 0.15));

                    namespace.SoundManager.playGunSound();
                    break;
                case namespace.WeaponTypeEnum.ShotGun:
                    firingCooldown = 0.7;
                    damage = (closestEntity.distance > 4.5) ? 1 : 3;
                    damage *= damageMultiplier;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.MultiBullet, damage));
                    levelManager.effects.push(new namespace.SmallSmokeEffect(position.x, position.y, 0.15));

                    namespace.SoundManager.playGunSound();
                    break;
                case namespace.WeaponTypeEnum.Uzi:
                    firingCooldown = 0.14;
                    damage = 0.7;
                    damage *= damageMultiplier;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.Bullet, damage));

                    namespace.SoundManager.playMachineGunSound();
                    break;
                case namespace.WeaponTypeEnum.AutoRifle:
                    firingCooldown = 0.18;
                    damage = 1;
                    damage *= damageMultiplier;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.Bullet, damage));

                    namespace.SoundManager.playMachineGunSound();
                    break;
                case namespace.WeaponTypeEnum.AutoShotGun:
                    firingCooldown = 0.2;
                    damage = (closestEntity.distance > 4.5) ? 0.45 : 1.5;
                    damage *= damageMultiplier;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.MultiBullet, damage));

                    namespace.SoundManager.playMachineGunSound();
                    break;
                case namespace.WeaponTypeEnum.Rpg:
                    firingCooldown = 1.7;
                    damage = 0;
                    levelManager.effects.push(new namespace.ProjectileEffect(position.x, position.y - 0.2, 0.15, angle, 11 * factorX, 11 * factorY, namespace.EffectTypeEnum.Warhead, damage));
                    levelManager.effects.push(new namespace.SmallSmokeEffect(position.x, position.y, 0.15));

                    effect = new namespace.BigSmokeEffect(position.x, position.y, 0.15);
                    effect.selfDestroyTime = 0.4;
                    effect.startFadeTime = 0.5;
                    levelManager.effects.push(effect);

                    namespace.SoundManager.playBlastSound();
                    break;
            }

            // Add the projectile effect.
            //levelManager.effects.push(effect);

            levelManager.effects.push(new namespace.SmallSmokeEffect(position.x - 0.1, position.y, 0.15));
            levelManager.effects.push(new namespace.SmallSmokeEffect(position.x + 0.1, position.y, 0.15));

            if (player.weapons[player.currentWeaponCategory] !== namespace.WeaponTypeEnum.Rpg && (Math.random() > 0.5 || player.weapons[player.currentWeaponCategory] === namespace.WeaponTypeEnum.ShotGun || player.weapons[player.currentWeaponCategory] === namespace.WeaponTypeEnum.Rifle)) {
                levelManager.effects.push(new namespace.CartridgeCaseEffect(player.x, player.y - 0.2, 0.15, (player.direction === namespace.DirectionEnum.Right) ? namespace.DirectionEnum.Left : namespace.DirectionEnum.Right));
            }

            // Add some smoke effect.

            // levelManager.effects.push(new namespace.SmallSmokeEffect(position.x, position.y, 0.15));
            // levelManager.effects.push(new namespace.SmallSmokeEffect(position.x + 0.1, position.y, 0.15));
            // levelManager.effects.push(new namespace.SmallSmokeEffect(position.x - 0.1, position.y - 0.15, 0.15));

            player.firingAngle = angle;
            player.firingCooldown = firingCooldown + player.firingTimeleftOver;
        }
    }

    function createDyingEntityEffect(x, y, direction) {
        var speed = 2 + Math.random() * 2;
        var effect = new namespace.Effect(x, y, 0.15);
        effect.mass = 0.7;
        effect.type = namespace.EffectTypeEnum.HeadLimb;
        effect.value = 0;
        effect.angle = Math.random();
        effect.rotationSpeed = (0.6 + Math.random() * 0.15) * Math.PI;
        effect.restoreBaseAngle = true;
        effect.direction = direction;
        if (cameraShakingTime > 0) {
            effect.selfDestroyTime = 2;
            effect.friction = 1.5;
            effect.vy = -speed * 2.2;
            effect.vx = ((Math.random() > 0.5) ? speed : -speed);
        }
        else {
            effect.selfDestroyTime = 2;
            effect.friction = 3;

            effect.vy = -speed;
            effect.vx = ((direction === namespace.DirectionEnum.Left) ? speed * 1.2 : -speed * 1.2);

            if (levelManager.currentLevel === 1 && levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWave) {
                effect.vy *= 1.7;
            }
        }
        levelManager.effects.push(effect);

        speed = 2 + Math.random();
        effect = new namespace.Effect(x, y, 0.15);
        effect.friction = 1;
        effect.mass = 0.7;
        effect.selfDestroyTime = 2;
        effect.type = namespace.EffectTypeEnum.BodyLimb1;
        effect.value = 0;
        effect.angle = Math.random();
        effect.rotationSpeed = (0.6 + Math.random() * 0.15) * Math.PI;
        effect.vx = ((Math.random() > 0.5) ? speed : -speed);
        effect.vy = -speed * 2;
        effect.direction = direction;
        levelManager.effects.push(effect);

        speed = 1.5 + Math.random();
        effect = new namespace.Effect(x, y, 0.15);
        effect.mass = 0.7;
        effect.selfDestroyTime = 1.8;
        effect.type = namespace.EffectTypeEnum.BodyLimb2;
        effect.value = 0;
        effect.angle = Math.random();
        effect.rotationSpeed = (0.6 + Math.random() * 0.15) * Math.PI;
        effect.vx = ((Math.random() > 0.5) ? speed : -speed);
        effect.direction = direction;
        if (cameraShakingTime > 0) {
            effect.friction = 1;
            effect.vy = -speed * 2;
        }
        else {
            effect.friction = 2;
            effect.vy = -speed;
        }
        levelManager.effects.push(effect);
        savedData.gold += 2;

        namespace.SoundManager.playZombieDieSound();
    }

    function addGainSurvivalPointEffect(value, x, y) {
        var effect = new namespace.ProjectileEffect(x, y, 0.5, 0, 0, -3, ((value === 10) ? namespace.EffectTypeEnum.GainSP : namespace.EffectTypeEnum.GainMoreSP), 0);
        effect.selfDestroyTime = 1;
        effect.startFadeTime = 0.7;
        levelManager.effects.push(effect);

        levelManager.survivalPointsZoomTime = 0.3;
    }

    function addStackObject(type, player) {
        var stackFound, u, midX, limitX, stack;
        if (type !== namespace.StackObjectTypeEnum.None) {
            if (player.buildDirection === namespace.DirectionEnum.Left) {
                if (type === namespace.StackObjectTypeEnum.WoodCrate || type === namespace.StackObjectTypeEnum.MetalCrate) {
                    stackFound = false;
                    for (u = 0; u < levelManager.stacks.length; ++u) {
                        if (levelManager.leftPlayer.stack === levelManager.stacks[u]) {
                            stackFound = true;
                            break;
                        }
                    }
                    if (!stackFound) {
                        levelManager.stacks.push(levelManager.leftPlayer.stack);
                    }
                    levelManager.leftPlayer.stack.objects.push(new namespace.StackObject(levelManager.maxY, 0.5, type, levelManager.getStackObjectStrength(type)));
                    levelManager.survivalPoints += 10;
                    addGainSurvivalPointEffect(10, levelManager.leftPlayer.stack.x, levelManager.maxY - 1.3);
                    namespace.SoundManager.playHumanSuccess();

                    levelManager.metalCount += (type === namespace.StackObjectTypeEnum.MetalCrate) ? 1 : 0;
                }
                else {
                    midX = (levelManager.minX + (levelManager.maxX - levelManager.minX) / 2);
                    limitX = midX - 1.5;
                    stack = { "x": limitX, "vx": -7, "objects": [], "limitX": limitX, "direction": namespace.DirectionEnum.Left, "cooldown": 1 };

                    if (type === namespace.StackObjectTypeEnum.DoubleWoodCrate) {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.WoodCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.WoodCrate)));
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.WoodCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.WoodCrate)));
                    }
                    else if (type === namespace.StackObjectTypeEnum.DoubleMetalCrate) {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.MetalCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.MetalCrate)));
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.MetalCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.MetalCrate)));
                    }
                    else {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, type, levelManager.getStackObjectStrength(type)));
                    }

                    stack.objects[0].rotationSpeed = 0.4 + Math.random() * 0.1;
                    levelManager.stacks.push(stack);
                    levelManager.survivalPoints += 10;
                    addGainSurvivalPointEffect(10, limitX, levelManager.maxY - 1.3);
                    namespace.SoundManager.playHumanSuccess();

                    levelManager.usedSpiked |= (type === namespace.StackObjectTypeEnum.SpikeCrate);
                    levelManager.usedTurret |= (type === namespace.StackObjectTypeEnum.MountedCrate);
                    levelManager.TNTCount += (type === namespace.StackObjectTypeEnum.TNTCrate) ? 1 : 0;
                }
            }
            else {
                if (type === namespace.StackObjectTypeEnum.WoodCrate || type === namespace.StackObjectTypeEnum.MetalCrate) {
                    stackFound = false;
                    for (u = 0; u < levelManager.stacks.length; ++u) {
                        if (levelManager.rightPlayer.stack === levelManager.stacks[u]) {
                            stackFound = true;
                            break;
                        }
                    }
                    if (!stackFound) {
                        levelManager.stacks.push(levelManager.rightPlayer.stack);
                    }
                    levelManager.rightPlayer.stack.objects.push(new namespace.StackObject(levelManager.maxY, 0.5, type, levelManager.getStackObjectStrength(type)));
                    levelManager.survivalPoints += 10;
                    addGainSurvivalPointEffect(10, levelManager.rightPlayer.stack.x, levelManager.maxY - 1.3);
                    namespace.SoundManager.playHumanSuccess();

                    levelManager.metalCount += (type === namespace.StackObjectTypeEnum.MetalCrate) ? 1 : 0;
                }
                else {
                    midX = (levelManager.minX + (levelManager.maxX - levelManager.minX) / 2);
                    limitX = midX + 1.5;
                    stack = { "x": limitX, "vx": 7, "objects": [], "limitX": limitX, "direction": namespace.DirectionEnum.Right, "cooldown": 1 };

                    if (type === namespace.StackObjectTypeEnum.DoubleWoodCrate) {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.WoodCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.WoodCrate)));
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.WoodCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.WoodCrate)));
                    }
                    else if (type === namespace.StackObjectTypeEnum.DoubleMetalCrate) {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.MetalCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.MetalCrate)));
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, namespace.StackObjectTypeEnum.MetalCrate, levelManager.getStackObjectStrength(namespace.StackObjectTypeEnum.MetalCrate)));
                    }
                    else {
                        stack.objects.push(new namespace.StackObject(levelManager.maxY - 1.3, 0.5, type, levelManager.getStackObjectStrength(type)));
                    }
                    stack.objects[0].rotationSpeed = 0.4 + Math.random() * 0.1;
                    levelManager.stacks.push(stack);
                    levelManager.survivalPoints += 10;
                    addGainSurvivalPointEffect(10, limitX, levelManager.maxY - 1.3);
                    namespace.SoundManager.playHumanSuccess();

                    levelManager.usedSpiked |= (type === namespace.StackObjectTypeEnum.SpikeCrate);
                    levelManager.usedTurret |= (type === namespace.StackObjectTypeEnum.MountedCrate);
                    levelManager.TNTCount += (type === namespace.StackObjectTypeEnum.TNTCrate) ? 1 : 0;
                }
            }
        }

        switch (levelManager.currentLevel) {
            case 1:
                if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialBuildingFirstCrate) {
                    levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialFirstWave;
                }
                else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialReinforceDefenses) {
                    levelManager.tutorialCooldown = 4.5;
                    levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialMoreZombiesComing;
                }
                else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamWork) {
                    if (levelManager.leftPlayer.stack.objects.length > 1 && levelManager.rightPlayer.stack.objects.length > 0) {
                        levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialThirdWave;
                    }
                }
                break;
            default:
                break;
        }

        if (levelManager.reachedTop === false) {
            if (levelManager.leftPlayer.stack.objects.length === 3 || levelManager.rightPlayer.stack.objects.length === 3) {
                levelManager.reachedTop = true;
            }
        }
    }

    function showReminder() {
        if (!levelManager.showReminder) {
            levelManager.showReminder = true;
            levelManager.showReminderTime = 4;
            levelManager.reachedTop = false;
        }
    }

    function updateEntities(elapsed) {
        var effect1;
        if (currentGameState !== namespace.GameStateEnum.GameLost) {
            var i, j, y, u;
            for (i = 0; i < levelManager.entities.length;) {
                var entity = levelManager.entities[i];
                if (entity.isDead) {
                    // Get rid of the entity.
                    createDyingEntityEffect(entity.x, entity.y, entity.direction);

                    levelManager.killCount++;
                    if (levelManager.currentLevel === 1000 || (levelManager.currentLevel >= 21 && levelManager.currentLevel <= 25)) {
                        if (levelManager.showHighScoreEffect && levelManager.killCount > 0 && levelManager.killCount > savedData.killCountHighScore) {
                            // Checkpoint effect.
                            effect1 = new namespace.ProjectileEffect(levelManager.leftPlayer.x + (levelManager.rightPlayer.x - levelManager.leftPlayer.x) / 2, levelManager.maxY - 1, 0.5, 0, 0, -3, namespace.EffectTypeEnum.HighScore, 0);
                            effect1.selfDestroyTime = 1.5;
                            effect1.startFadeTime = 0.7;
                            levelManager.effects.push(effect1);
                            namespace.SoundManager.playHighScoreSound();
                            levelManager.showHighScoreEffect = false;
                        }
                        savedData.killCountHighScore = Math.max(savedData.killCountHighScore, levelManager.killCount);

                        if ((levelManager.killCount % levelManager.checkpointFrequency) === 0) {
                            // Checkpoint effect.
                            effect1 = new namespace.ProjectileEffect(levelManager.leftPlayer.x + (levelManager.rightPlayer.x - levelManager.leftPlayer.x) / 2, levelManager.maxY - 1, 0.5, 0, 0, -3, namespace.EffectTypeEnum.Checkpoint, 0);
                            effect1.selfDestroyTime = 1.5;
                            effect1.startFadeTime = 0.7;
                            levelManager.effects.push(effect1);
                            namespace.SoundManager.playCheckpointSound();
                        }
                    }

                    levelManager.survivalPoints += (entity.bonus ? 15 : 10);
                    addGainSurvivalPointEffect(entity.bonus ? 15 : 10, entity.x, entity.y - 0.4);

                    switch (levelManager.currentLevel) {
                        case 1:
                            if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWave) {
                                levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialFirstKill;
                                levelManager.tutorialCooldown = 0.35;
                            }
                            break;
                        default:
                            break;
                    }

                    levelManager.entities.splice(i, 1);
                    continue;
                }

                entity.update(elapsed);

                // Make sure the entity is always above ground level.
                y = levelManager.maxY;
                if (entity.y >= y - entity.radius) {
                    entity.y = y - entity.radius;
                    entity.vy = 0;
                }

                // Determine if the entity is visible on screen.
                var limit1 = namespace.screenToScene({ "x": 0, "y": 0 });
                var limit2 = namespace.screenToScene({ "x": namespace.getCanvasWidth(), "y": 0 });
                entity.offScreenDirection = (entity.x >= limit1.x && entity.x <= limit2.x) ? namespace.DirectionEnum.None : ((entity.x < limit1.x) ? namespace.DirectionEnum.Left : namespace.DirectionEnum.Right);

                if (entity.direction !== namespace.DirectionEnum.None) {
                    // Get the closest stack ahead.
                    var closestStack = null;
                    var currentDistance = levelManager.maxX - levelManager.minX, distance = 0;
                    for (j = 0; j < levelManager.stacks.length; ++j) {
                        var stack = levelManager.stacks[j];
                        if ((entity.direction === namespace.DirectionEnum.Left && stack.x < entity.x) || (entity.direction === namespace.DirectionEnum.Right && stack.x > entity.x)) {
                            distance = (entity.direction === namespace.DirectionEnum.Left) ? entity.x - stack.x : stack.x - entity.x;
                            if (closestStack === null || distance < currentDistance) {
                                if (stack.objects.length > 0) {
                                    currentDistance = distance;
                                    closestStack = stack;
                                }
                            }
                        }
                    }

                    // Close enough to start charging.
                    if (!entity.bouncing && closestStack !== null && entity.moveCooldown === 0 && currentDistance < 1.5 + entity.radius) {
                        entity.speed = entity.baseSpeed * 14;
                        entity.rotationSpeed = 5;

                        // If the charge was started from appropriate distance we can register the damage.
                        if (!entity.canApplyDamage && currentDistance >= 1) {
                            entity.canApplyDamage = true;
                            namespace.SoundManager.playZombieAttackSound();
                        }

                        // Bounce.
                        if (currentDistance < entity.radius + 0.5) {
                            entity.rotationSpeed = 0;
                            entity.vy -= 2;
                            entity.vx = (entity.direction === namespace.DirectionEnum.Left) ? 4.2 + Math.random() : -4.2 - Math.random();
                            entity.bouncing = true;

                            if (cameraShakingTime === 0 && closestStack.direction === namespace.DirectionEnum.None && cameraShakingTime === 0 && slowMotionTime === 0) {
                                cameraShakingDir = (Math.random() > 0.5) ? namespace.DirectionEnum.Left : namespace.DirectionEnum.Right;
                                cameraShakingTime = 0.5;
                                wantSlowMotion = false;
                            }

                            switch (closestStack.objects[0].type) {
                                case namespace.StackObjectTypeEnum.WoodCrate:
                                case namespace.StackObjectTypeEnum.TNTCrate:
                                case namespace.StackObjectTypeEnum.MountedCrate:
                                    entity.hitPoints -= 3 * entity.armorCoefficient;
                                    break;
                                case namespace.StackObjectTypeEnum.SpikeCrate:
                                    entity.hitPoints -= 5;
                                    break;
                                case namespace.StackObjectTypeEnum.MetalCrate:
                                    entity.hitPoints -= 5 * entity.armorCoefficient;
                                    break;
                            }

                            if (closestStack.direction !== namespace.DirectionEnum.None) {
                                closestStack.vx = closestStack.vx + ((entity.direction === namespace.DirectionEnum.Left) ? -1.5 : 1.5);
                            }

                            if (entity.canApplyDamage) {
                                entity.canApplyDamage = false;

                                if (closestStack.cooldown === 0) {
                                    closestStack.objects[0].strength -= 1 * entity.damageFactor;

                                    if (closestStack.objects[0].type === namespace.StackObjectTypeEnum.WoodCrate) {
                                        levelManager.noScratch = false;
                                    }
                                }

                                if (closestStack.cooldown === 0 && closestStack.objects[0].vy >= -0.5 && closestStack.objects[0].vy <= 0.5) {
                                    closestStack.objects[0].vy = -1.5;
                                }

                                // Simulate force transfer up the stack.
                                for (u = 1; u < closestStack.objects.length; ++u) {
                                    if (closestStack.objects[u].vy === 0) {
                                        closestStack.objects[u].vy = -2.2;
                                    }
                                }

                                if (closestStack === levelManager.leftPlayer.stack && levelManager.leftPlayer.vy === 0) {
                                    levelManager.leftPlayer.vy = -3.5;
                                }
                                else if (closestStack === levelManager.rightPlayer.stack && levelManager.rightPlayer.vy === 0) {
                                    levelManager.rightPlayer.vy = -3.5;
                                }
                            }
                        }
                    }
                    else {
                        if (closestStack !== null && currentDistance < entity.radius + 1) {
                            //  closestStack.vx = 0;
                            entity.x = (entity.direction === namespace.DirectionEnum.Left) ? closestStack.x + (entity.radius + 1) : closestStack.x - (entity.radius + 1);
                        }
                        entity.speed = entity.baseSpeed;
                        entity.rotationSpeed = 0;
                        entity.canApplyDamage = false;
                    }

                    if (entity.bouncing) {
                        entity.woundedCooldown = 0.1;
                    }

                    if (entity.direction === namespace.DirectionEnum.Right && entity.angle > namespace.toRadians(55)) {
                        entity.angle = namespace.toRadians(55);
                    }
                    else if (entity.direction === namespace.DirectionEnum.Left && entity.angle > namespace.toRadians(55)) {
                        entity.angle = namespace.toRadians(55);
                    }
                }

                if (entity.x < levelManager.minX && entity.direction === namespace.DirectionEnum.Left) {
                    entity.direction = namespace.DirectionEnum.Right;
                }
                else if (entity.x > levelManager.maxX && entity.direction === namespace.DirectionEnum.Right) {
                    entity.direction = namespace.DirectionEnum.Left;
                }

                // Check for collisions with the player.
                if (namespace.rectIntersect(levelManager.leftPlayer.x - levelManager.leftPlayer.radius, levelManager.leftPlayer.y - levelManager.leftPlayer.radius, levelManager.leftPlayer.x + levelManager.leftPlayer.radius, levelManager.leftPlayer.y + levelManager.leftPlayer.radius, entity.x - entity.radius, entity.y - entity.radius, entity.x + entity.radius, entity.y + entity.radius) ||
                    namespace.rectIntersect(levelManager.rightPlayer.x - levelManager.rightPlayer.radius, levelManager.rightPlayer.y - levelManager.rightPlayer.radius, levelManager.rightPlayer.x + levelManager.rightPlayer.radius, levelManager.rightPlayer.y + levelManager.rightPlayer.radius, entity.x - entity.radius, entity.y - entity.radius, entity.x + entity.radius, entity.y + entity.radius)) {
                    if (currentGameState !== namespace.GameStateEnum.GameLost) {
                        namespace.SoundManager.playHumanDie();
                        savedData.lives--;
                        if (savedData.lives < 0) {
                            savedData.lives = 0;
                        }
                        showReviveScreen = true;
                        currentGameState = namespace.GameStateEnum.GameLost;
                        levelManager.levelEndCooldown = 0.3;
                        levelManager.levelEndPauseCooldown = 1;
                        if (entity.direction === namespace.DirectionEnum.Right) {
                            levelManager.leftPlayer.isDead = true;
                            levelManager.leftPlayer.firingCooldown = 0;
                            levelManager.leftPlayer.vy = -5.5;
                            levelManager.leftPlayer.vx = 5;
                        }
                        else {
                            levelManager.rightPlayer.isDead = true;
                            levelManager.rightPlayer.firingCooldown = 0;
                            levelManager.rightPlayer.vy = -5.5;
                            levelManager.rightPlayer.vx = -5;
                            levelManager.stacks = [];
                        }
                        levelHudItems = [];

                        for (u = 0; u < levelManager.entities.length; u += 1) {
                            levelManager.entities[u].drawHitPoints = false;
                        }

                        // Save the data.
                        namespace.saveData();
                    }
                }

                i += 1;
            }
        }
    }

    function getClosestEntity(x, y, direction) {
        var i, len, entity = null, currentEntity = null, distance = levelManager.maxX - levelManager.minX, currentDistance = 0;
        for (i = 0, len = levelManager.entities.length; i < len; i += 1) {
            currentEntity = levelManager.entities[i];
            if ((direction === namespace.DirectionEnum.Left && currentEntity.x < x) || (direction === namespace.DirectionEnum.Right && currentEntity.x > x)) {
                if (currentEntity.direction === namespace.DirectionEnum.Left && currentEntity.x < levelManager.maxX || currentEntity.direction === namespace.DirectionEnum.Right && currentEntity.x > levelManager.minX) {
                    currentDistance = namespace.distance(x, y, currentEntity.x, currentEntity.y);
                    if (currentDistance < distance) {
                        entity = currentEntity;
                        distance = currentDistance;
                    }
                }
            }
        }
        return { "entity": entity, "distance": distance };
    }

    function createConfirmDialog() {
        var size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
        levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ConfirmNo, "x": size.x / 2 - 1.25, "y": size.y / 2 - 0.5, "size": 0.95 });
        levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ConfirmYes, "x": size.x / 2 + 0.25, "y": size.y / 2 - 0.5, "size": 0.95 });
    }

    function showIAPScreen() {
        namespace.SoundManager.playReloadSound();
        levelManager.showProducts = true;
        levelManager.productTime = 0.5;
    }

    function restartCurrentLevel(fullStack) {
        if (levelManager) {
            levelManager.entities = [];
            currentGameState = namespace.GameStateEnum.GameRun;
            gamePaused = false;

            createHudItems(levelManager.currentLevel);
            savedData.lives++;
            levelManager.levelEndCooldown = 0.3;
            levelManager.levelEndPauseCooldown = 1;

            var positionX = levelManager.minX + (levelManager.maxX - levelManager.minX) / 2;
            levelManager.rightPlayer.x = positionX + 0.6;
            levelManager.leftPlayer.x = positionX - 0.6;
            levelManager.leftPlayer.vx = 0;
            levelManager.rightPlayer.vx = 0;
            levelManager.leftPlayer.isDead = false;
            levelManager.rightPlayer.isDead = false;

            if (!fullStack) {
                levelManager.survivalPoints += 100;
            }
            else {
                levelManager.leftPlayer.stack.objects = [];
                levelManager.rightPlayer.stack.objects = [];
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
            }

            namespace.SoundManager.playAward();
        }
    }

    function pointerDown(event) {
        var maxCount, i, u, size, buildData, currentType, clickedHandled = false;

        if (!desktop) {
            var x = (event.targetTouches[0].pageX - canvas.offsetLeft);
            var y = (event.targetTouches[0].pageY - canvas.offsetTop);
            var point = namespace.screenToScene({ "x": x / optimalRatioX, "y": y / optimalRatioY });
            var hudPoint = namespace.screenToHud({ "x": x / optimalRatioX, "y": y / optimalRatioY });
        }
        else {
            var x = event.pageX - canvas.offsetLeft;
            var y = event.pageY - canvas.offsetTop;
            var point = namespace.screenToScene({ "x": x / optimalRatioX, "y": y / optimalRatioY });
            var hudPoint = namespace.screenToHud({ "x": x / optimalRatioX, "y": y / optimalRatioY });
        }

        // Remove canvas offset.
       // var x = (event.targetTouches[0].pageX - canvas.offsetLeft);
       // var y = (event.targetTouches[0].pageY - canvas.offsetTop);

        // Project the points to appropriate cameras.
        //var point = namespace.screenToScene({ "x": x / optimalRatioX, "y": y / optimalRatioY });
        //var hudPoint = namespace.screenToHud({ "x": x / optimalRatioX, "y": y / optimalRatioY });

       // var point = namespace.screenToScene({ "x": x, "y": y });
       // var hudPoint = namespace.screenToHud({ "x": x, "y": y });

        isPointerDown = true;

        if (levelManager && levelManager.showProducts) {
            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

            if (namespace.pointInRect(hudPoint.x, hudPoint.y, 1.3, 0.4, 1.3 + 2.5, 0.4 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("001");
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 3.9, 0.4, 3.9 + 2.5, 0.4 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("002");
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 6.5, 0.4, 6.5 + 2.5, 0.4 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("003");
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 1.3, 3.1, 1.3 + 2.5, 3.1 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("004");
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 3.9, 3.1, 3.9 + 2.5, 3.1 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("005");
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 6.5, 3.1, 6.5 + 2.5, 3.1 + 2.6)) {
                namespace.SoundManager.playReloadSound();
                namespace.buyItem("006");
            }
            else {
                levelManager.showProducts = false;
                gamePaused = false;
            }
        }
        else if (showReviveScreen) {
            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

            // Click cancel button, do nothing.
            if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x / 2 + 2.5, size.y / 2 - 2.5, size.x / 2 + 3.5, size.y / 2 - 1.5)) {
                showReviveScreen = false;
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x / 2 - 3, size.y / 2 - 1.5, size.x / 2 - 0.5, size.y / 2 + 2.5)) {
                if (savedData.gold >= 500) {
                    savedData.gold -= 500;
                    namespace.saveData();
                    showReviveScreen = false;
                    restartCurrentLevel(true);
                }
                else {
                    levelManager.showProducts = true;
                    levelManager.productTime = 0;
                }
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x / 2 + 0.5, size.y / 2 - 1.5, size.x / 2 + 3, size.y / 2 + 2.5)) {
                if (savedData.gold >= 750) {
                    savedData.gold -= 750;
                    namespace.saveData();
                    showReviveScreen = false;
                    restartCurrentLevel(true);
                }
                else {
                    levelManager.showProducts = true;
                    levelManager.productTime = 0;
                }
            }
        }
        else if (levelManager && levelManager.showBoosters) {
            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

            // Gold button.
            if (namespace.pointInRect(hudPoint.x, hudPoint.y, 0.4, 0.7, 0.4 + 2.4, 0.7 + 2)) {
                showIAPScreen();
            }
            else if (savedData.lives < 3 && namespace.pointInRect(hudPoint.x, hudPoint.y, 0.4, 3.7, 0.4 + 2.4, 3.7 + 2)) {
                if (savedData.gold >= 2000) {
                    savedData.gold -= 2000;
                    savedData.lives = 3;
                    savedData.refillTime = 0;
                    namespace.SoundManager.playAward();
                    namespace.saveData();
                }
                else {
                    showIAPScreen();
                }
            }
            else if (!levelManager.boosters[0].activated && namespace.pointInRect(hudPoint.x, hudPoint.y, 3.2, 0, 3.2 + 3, 3)) {
                if (savedData.gold >= 500 || levelManager.boosters[0].free) {
                    if (!levelManager.boosters[0].free) {
                        savedData.gold -= 500;
                    }
                    levelManager.boosters[0].activated = true;
                    namespace.SoundManager.playBlastSound();
                }
                else {
                    showIAPScreen();
                }
            }
            else if (!levelManager.boosters[1].activated && namespace.pointInRect(hudPoint.x, hudPoint.y, 6.3, 0, 6.3 + 3, 3)) {
                if (savedData.gold >= 750 || levelManager.boosters[1].free) {
                    if (!levelManager.boosters[1].free) {
                        savedData.gold -= 750;
                    }
                    levelManager.boosters[1].activated = true;
                    namespace.SoundManager.playBlastSound();
                }
                else {
                    showIAPScreen();
                }
            }
            else if (!levelManager.boosters[2].activated && namespace.pointInRect(hudPoint.x, hudPoint.y, 3.2, 3.2, 3.2 + 3, 3.2 + 3)) {
                if (savedData.gold >= 1500 || levelManager.boosters[2].free) {
                    if (!levelManager.boosters[2].free) {
                        savedData.gold -= 1500;
                    }
                    levelManager.boosters[2].activated = true;
                    namespace.SoundManager.playBlastSound();
                }
                else {
                    showIAPScreen();
                }
            }
            else if (!levelManager.boosters[3].activated && namespace.pointInRect(hudPoint.x, hudPoint.y, 6.3, 3.2, 6.3 + 3, 3.2 + 3)) {
                if (savedData.gold >= 2500 || levelManager.boosters[3].free) {
                    if (!levelManager.boosters[3].free) {
                        savedData.gold -= 2500;
                    }
                    levelManager.boosters[3].activated = true;
                    namespace.SoundManager.playBlastSound();
                }
                else {
                    showIAPScreen();
                }
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, 9.3, 2.3, 9.3 + 1.5, 2.3 + 1.5)) {

                if (savedData.lives === 0) {
                    levelManager.refillMessageTime = 3;
                    namespace.SoundManager.playReloadSound();
                }
                else {
                    levelManager.showBoosters = false;
                    namespace.SoundManager.playCritSound();

                    // Update hud level items.
                    if (levelManager.currentLevel === 1000 || levelManager.currentLevel === 1001 || (levelManager.currentLevel >= 21 && levelManager.currentLevel <= 25)) {
                        var item, i, len = levelItemData.length;
                        var level = 6;
                        for (i = 0; i < len; i += 1) {
                            item = levelItemData[i];
                            if (item.status !== namespace.LevelStatusEnum.Passed) {
                                break;
                            }
                            level = Math.max(level, i + 1);
                        }
                        createHudItems(level);
                    }
                    else {
                        createHudItems();
                    }
                }
            }
        }
        else if (showReviewMsg) {
            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
            if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x / 2 - 2.5, size.y - 3, size.x / 2 - 2.5 + 2, size.y - 3 + 2)) {
                showReviewMsg = false;
                showThanksMsgTime = 3;
            }
            else if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x / 2 + 0.5, size.y - 3, size.x / 2 + 0.5 + 2, size.y - 3 + 2)) {
                showReviewMsg = false;
                if (typeof Cocoon !== "undefined" && Cocoon.App) {
                    Cocoon.App.openURL("https://play.google.com/store/apps/details?id=com.hopy.zombies.td");
                }
            }
        }
        else if (currentGameState === namespace.GameStateEnum.GameIntro) {
            gameIntroTime = Math.min(gameIntroTime, 1);
        }
        else if (confirmCommand !== namespace.CommandEnum.None) {
            for (i = 0; i < levelHudItems.length; i += 1) {
                if (levelHudItems[i].enabled && namespace.pointInRect(hudPoint.x, hudPoint.y, levelHudItems[i].x, levelHudItems[i].y, levelHudItems[i].x + levelHudItems[i].size, levelHudItems[i].y + levelHudItems[i].size)) {
                    currentType = levelHudItems[i].type;
                    if (currentType === namespace.CommandEnum.ConfirmYes) {
                        switch (confirmCommand) {
                            case namespace.CommandEnum.Menu:
                                namespace.showMainScreen();
                                gamePaused = false;
                                confirmCommand = namespace.CommandEnum.None;
                                clickedHandled = true;
                                showFullScreenAd();
                                break;
                            case namespace.CommandEnum.Restart:
                                loadLevel(levelManager.currentLevel);
                                gamePaused = false;
                                confirmCommand = namespace.CommandEnum.None;
                                clickedHandled = true;
                                break;
                        }
                        break;
                    }
                    else if (currentType === namespace.CommandEnum.ConfirmNo) {
                        clickedHandled = true;
                        namespace.cancelConfirmDialog();
                        break;
                    }
                }
            }
        }
        else {
            switch (currentGameState) {
                case namespace.GameStateEnum.GameMenu:
                    if (showCredits) {
                        showCredits = false;
                    }
                    else if (!(currentMenuType === namespace.MenuTypeEnum.None && gameMenuSetupTime > 2)) {
                        var handled = false;

                        size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                        if (namespace.isTrialLicense()) {
                            if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x - 2, size.y - 1, size.x, size.y)) {
                                loadLevel(6);
                                handled = true;
                            }
                        }

                        if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x - 1, size.y - 1, size.x, size.y)) {
                            if (typeof Cocoon !== "undefined" && Cocoon.App) {
                                Cocoon.App.openURL("http://www.okijin.com");
                            }
                            handled = true;
                        }

                        if (currentMenuType === namespace.MenuTypeEnum.ShowStoryMode) {
                            for (i = menuPageIndex * 9; i < Math.min(levelItemData.length, (menuPageIndex + 1) * 9) ; i += 1) {
                                if (namespace.pointInRect(hudPoint.x, hudPoint.y, levelItemData[i].x, levelItemData[i].y, levelItemData[i].x + levelItemData[i].size, levelItemData[i].y + levelItemData[i].size)) {
                                    if (levelItemData[i].status === namespace.LevelStatusEnum.Available || levelItemData[i].status === namespace.LevelStatusEnum.Passed) {
                                        loadLevel(levelItemData[i].level);
                                    }
                                    else if (levelItemData[i].status === namespace.LevelStatusEnum.Locked) {
                                        itemLockedDialogTime = 1.7;
                                    }
                                    handled = true;
                                    break;
                                }
                            }
                        }
                        else if (currentMenuType === namespace.MenuTypeEnum.ShowSurvivalMode) {
                            if (namespace.pointInRect(hudPoint.x, hudPoint.y, menuSurvivalItems[menuPageIndex].x, menuSurvivalItems[menuPageIndex].y, menuSurvivalItems[menuPageIndex].x + menuSurvivalItems[menuPageIndex].size, menuSurvivalItems[menuPageIndex].y + menuSurvivalItems[menuPageIndex].size)) {
                                if (menuSurvivalItems[menuPageIndex].status === namespace.LevelStatusEnum.Available || menuSurvivalItems[menuPageIndex].status === namespace.LevelStatusEnum.Passed) {
                                    loadLevel(menuSurvivalItems[menuPageIndex].level);
                                }
                                else if (menuSurvivalItems[menuPageIndex].status === namespace.LevelStatusEnum.Locked) {
                                    itemLockedDialogTime = 1.7;
                                }
                                handled = true;
                            }
                        }
                        else if (currentMenuType === namespace.MenuTypeEnum.ShowOptions) {
                            for (i = 0; i < menuAchievementItems.length; i += 1) {
                                if (namespace.pointInRect(hudPoint.x, hudPoint.y, menuAchievementItems[i].x, menuAchievementItems[i].y, menuAchievementItems[i].x + menuAchievementItems[i].size, menuAchievementItems[i].y + menuAchievementItems[i].size)) {
                                    handled = true;
                                    break;
                                }
                            }
                        }

                        if (!handled) {
                            for (i = 0; i < menuHudItems.length; i += 1) {
                                if (namespace.pointInRect(hudPoint.x, hudPoint.y, menuHudItems[i].x, menuHudItems[i].y, menuHudItems[i].x + menuHudItems[i].width, menuHudItems[i].y + menuHudItems[i].height)) {
                                    if (currentMenuType !== menuHudItems[i].type) {
                                        switch (menuHudItems[i].type) {
                                            case namespace.MenuTypeEnum.ToggleSound:
                                                if (namespace.SoundManager.getSoundsEnabled() === 0) {
                                                    namespace.SoundManager.unmute(true);
                                                }
                                                else if (namespace.SoundManager.getSoundsEnabled() === 2) {
                                                    namespace.SoundManager.mute(false);
                                                }
                                                else {
                                                    namespace.SoundManager.mute(true);
                                                }

                                                break;
                                            case namespace.MenuTypeEnum.ShowCredits:
                                                showCredits = true;
                                                break;
                                            case namespace.MenuTypeEnum.ShowMain:
                                            case namespace.MenuTypeEnum.ShowStoryMode:
                                            case namespace.MenuTypeEnum.ShowSurvivalMode:
                                            case namespace.MenuTypeEnum.ShowOptions:
                                                menuPageIndex = 0;
                                                gameMaxMenuSetupTime = 3;
                                                gameMenuSetupTime = gameMaxMenuSetupTime;
                                                currentMenuType = menuHudItems[i].type;
                                                menuPageTime = 0;
                                                menuPageDirection = namespace.DirectionEnum.None;
                                                namespace.SoundManager.playClickSound();
                                                break;
                                            case namespace.MenuTypeEnum.SelectNextPage:
                                                if (currentMenuType === namespace.MenuTypeEnum.ShowStoryMode) {
                                                    maxCount = Math.floor(levelItemData.length / 9);
                                                    menuPageIndex++;
                                                    if (menuPageIndex > maxCount) {
                                                        menuPageIndex = 0;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Right;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                else if (currentMenuType === namespace.MenuTypeEnum.ShowSurvivalMode) {
                                                    maxCount = menuSurvivalItems.length - 1;
                                                    menuPageIndex++;
                                                    if (menuPageIndex > maxCount) {
                                                        menuPageIndex = 0;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Right;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                else if (currentMenuType === namespace.MenuTypeEnum.ShowOptions) {
                                                    maxCount = menuAchievementItems.length - 1;
                                                    menuPageIndex++;
                                                    if (menuPageIndex > maxCount) {
                                                        menuPageIndex = 0;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Right;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                break;
                                            case namespace.MenuTypeEnum.SelectPrevPage:
                                                if (currentMenuType === namespace.MenuTypeEnum.ShowStoryMode) {
                                                    menuPageIndex--;
                                                    maxCount = Math.floor(levelItemData.length / 9);
                                                    if (menuPageIndex < 0) {
                                                        menuPageIndex = maxCount;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Left;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                else if (currentMenuType === namespace.MenuTypeEnum.ShowSurvivalMode) {
                                                    menuPageIndex--;
                                                    maxCount = menuSurvivalItems.length - 1;
                                                    if (menuPageIndex < 0) {
                                                        menuPageIndex = maxCount;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Left;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                else if (currentMenuType === namespace.MenuTypeEnum.ShowOptions) {
                                                    menuPageIndex--;
                                                    maxCount = menuAchievementItems.length - 1;
                                                    if (menuPageIndex < 0) {
                                                        menuPageIndex = maxCount;
                                                    }

                                                    menuPageTime = 0.25;
                                                    menuPageDirection = namespace.DirectionEnum.Left;
                                                    namespace.SoundManager.playClickSound();
                                                }
                                                break;
                                            default:
                                                break;
                                        }
                                    }
                                    break;
                                }
                            }
                        }
                    }
                    break;
                case namespace.GameStateEnum.GameLost:
                case namespace.GameStateEnum.GameWon:
                case namespace.GameStateEnum.GameRun:

                    if (levelManager && levelManager.showReminder) {
                        size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                        if (namespace.pointInRect(hudPoint.x, hudPoint.y, size.x - 5 * 0.7, size.y / 2 - 1.5, (size.x - 5 * 0.7) + 5 * 0.7, (size.y / 2 - 1.5) + 3 * 0.7)) {

                            gamePaused = true;

                            if (savedData.gold >= 300) {
                                savedData.gold -= 300;
                                levelManager.showReminder = false;

                                levelManager.leftPlayer.stack.objects = [];
                                levelManager.rightPlayer.stack.objects = [];
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.leftPlayer);
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
                                addStackObject(namespace.StackObjectTypeEnum.MetalCrate, levelManager.rightPlayer);
                                gamePaused = false;
                                namespace.SoundManager.playAward();
                            }
                            else {
                                levelManager.showProducts = true;
                                levelManager.productTime = 0;
                            }

                            return;
                        }
                    }

                    var playScreen = (levelManager.newItem === namespace.StackObjectTypeEnum.None && levelManager.primaryWeaponAssignmentDone && levelManager.secondaryWeaponAssignmentDone);

                    for (i = 0; i < levelHudItems.length; i += 1) {
                        if (levelHudItems[i].enabled && namespace.pointInRect(hudPoint.x, hudPoint.y, levelHudItems[i].x, levelHudItems[i].y, levelHudItems[i].x + levelHudItems[i].size, levelHudItems[i].y + levelHudItems[i].size)) {
                            currentType = levelHudItems[i].type;
                            if (currentType === namespace.CommandEnum.Menu) {
                                if (levelManager.currentLevel > 5 && namespace.isTrialLicense()) {
                                    namespace.showMainScreen();
                                }
                                else {
                                    namespace.displayMenuConfirmDialog();
                                }
                                clickedHandled = true;

                                namespace.SoundManager.playClickSound();
                            }
                            if (currentType === namespace.CommandEnum.NextLevel) {
                                canDisplayAd = true;
                                if ((levelManager.currentLevel === maxStoryModeLevel) || (levelManager.currentLevel === menuSurvivalItems[1].level)) {
                                    loadLevel(menuSurvivalItems[0].level);
                                }
                                else {
                                    loadLevel(levelManager.currentLevel + 1);
                                }
                                clickedHandled = true;
                                namespace.SoundManager.playReloadSound();
                            }
                            else if (currentType === namespace.CommandEnum.Restart) {
                                if (currentGameState !== namespace.GameStateEnum.GameLost) {
                                    confirmCommand = namespace.CommandEnum.Restart;
                                    createConfirmDialog();
                                    gameWasPaused = gamePaused;
                                    gamePaused = true;
                                }
                                else {
                                    loadLevel(levelManager.currentLevel);
                                }

                                clickedHandled = true;

                                namespace.SoundManager.playReloadSound();
                            }
                            else if (currentType === namespace.CommandEnum.Pause) {
                                gamePaused = !gamePaused;
                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.BuyGame) {
                                namespace.buyGame();
                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.ZoomIn) {
                                cameraTargetZoom += 0.05;
                                if (cameraTargetZoom > 0.9) {
                                    cameraTargetZoom = 0.9;
                                }
                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.ZoomOut) {
                                cameraTargetZoom -= 0.05;
                                if (cameraTargetZoom < 0.15) {
                                    cameraTargetZoom = 0.15;
                                }
                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.CrateSelectionDone) {
                                size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                                levelHudItems[i].ty = size.y - 1.1;
                                levelHudItems[i].y = size.y + 1;

                                if (levelManager.newItem !== namespace.StackObjectTypeEnum.None) {
                                    levelManager.newItem = namespace.StackObjectTypeEnum.None;
                                    addWeaponAssignmentItems(size);
                                }

                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.Ready) {

                                size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

                                levelHudItems[i].ty = size.y - 1.1;
                                levelHudItems[i].y = size.y + 1;

                                if (!levelManager.primaryWeaponAssignmentDone) {
                                    levelManager.primaryWeaponAssignmentDone = true;
                                }
                                else if (!levelManager.secondaryWeaponAssignmentDone) {
                                    levelManager.secondaryWeaponAssignmentDone = true;
                                }

                                if (levelManager.primaryWeaponAssignmentDone && levelManager.secondaryWeaponAssignmentDone) {
                                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Primary;
                                    levelManager.leftPlayer.direction = namespace.DirectionEnum.Left;
                                    levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
                                    levelHudItems[i].enabled = false;

                                    if (levelManager.currentLevel > 2) {
                                        showFullScreenAd(true);
                                    }

                                    //$("#canvas").css("background-image", "url('assets/img/683/background01.png')");
                                }
                                else if (levelManager.primaryWeaponAssignmentDone) {
                                    levelManager.leftPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                                    levelManager.rightPlayer.currentWeaponCategory = namespace.WeaponCategoryEnum.Secondary;
                                    levelManager.selectedWeaponType = namespace.WeaponTypeEnum.None;

                                    //$("#canvas").css("background-image", "url('assets/img/683/background02.png')");
                                }

                                levelManager.rightPlayer.targetScaleFactor = 1;
                                levelManager.leftPlayer.targetScaleFactor = 1;
                                clickedHandled = true;

                                namespace.SoundManager.playReloadSound();
                            }
                            else if (currentType === namespace.CommandEnum.ToggleAdrenaline && playScreen) {
                                if (levelManager.survivalRageTime === 0 && levelManager.survivalPoints >= levelManager.minSurvivalPoints) {
                                    var bonusFactor = 1;
                                    if (levelManager.leftPlayer.weapons[namespace.WeaponCategoryEnum.Secondary] === namespace.WeaponTypeEnum.Uzi) {
                                        bonusFactor += 0.15;
                                    }
                                    if (levelManager.rightPlayer.weapons[namespace.WeaponCategoryEnum.Secondary] === namespace.WeaponTypeEnum.Uzi) {
                                        bonusFactor += 0.15;
                                    }
                                    levelManager.survivalRageTime = levelManager.maxSurvivalRageTime * bonusFactor;
                                    levelManager.survivalPoints -= levelManager.minSurvivalPoints;
                                    levelManager.survivalRage = true;

                                    //$("#canvas").css("background-image", "url('assets/img/683/background02.png')");

                                    //clearHudSelection(true);

                                    switch (levelManager.currentLevel) {
                                        case 1:
                                            if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSelectSurvivalRage) {
                                                gamePaused = false;
                                                levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialFirstWaveFinishJob;
                                            }
                                            break;
                                        default:
                                            break;
                                    }
                                }
                                else if (levelManager.currentLevel > 2) {
                                    if (savedData.gold >= 100) {
                                        savedData.gold -= 100;
                                        levelManager.survivalPoints += 100;
                                        namespace.SoundManager.playAward();
                                    }
                                    else {
                                        showIAPScreen();
                                    }
                                }

                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.CancelAction && playScreen) {
                                if (levelHudItems[i].direction === namespace.DirectionEnum.Left && levelManager.leftPlayer.buildData.cooldown > 0) {
                                    levelManager.leftPlayer.buildData.cooldown = 0;
                                    levelManager.leftPlayer.buildData.stackObjectType = namespace.StackObjectTypeEnum.None;
                                    clickedHandled = true;
                                }
                                else if (levelHudItems[i].direction === namespace.DirectionEnum.Right && levelManager.rightPlayer.buildData.cooldown > 0) {
                                    levelManager.rightPlayer.buildData.cooldown = 0;
                                    levelManager.rightPlayer.buildData.stackObjectType = namespace.StackObjectTypeEnum.None;
                                    clickedHandled = true;
                                }

                                namespace.SoundManager.playClickSound();
                            }
                            else if (currentType === namespace.CommandEnum.SelectStackObject && playScreen) {
                                // Retrieve the selected ShowStackObjects.
                                var buildHudItem = null;
                                for (u = 0; u < levelHudItems.length; ++u) {
                                    if (levelHudItems[u].type === namespace.CommandEnum.ShowStackObjects) {
                                        if (levelHudItems[u].selected && levelHudItems[u].direction === levelHudItems[i].direction) {
                                            buildHudItem = levelHudItems[u];
                                            break;
                                        }
                                    }
                                }

                                var disabled = false;
                                if (buildHudItem) {
                                    if (levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.MetalCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.WoodCrate) {
                                        if (buildHudItem.direction === namespace.DirectionEnum.Left && levelManager.leftPlayer.stack.objects.length === levelManager.maxStackSize) {
                                            disabled = true;
                                        }
                                        else if (buildHudItem.direction === namespace.DirectionEnum.Right && levelManager.rightPlayer.stack.objects.length === levelManager.maxStackSize) {
                                            disabled = true;
                                        }
                                    }

                                    if (levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.RepairCrate) {
                                        var player = (levelHudItems[i].direction === namespace.DirectionEnum.Left) ? levelManager.leftPlayer : levelManager.rightPlayer;
                                        if (levelManager.repairKits <= 0 || (!(player.stack.objects.length > 0 && player.stack.objects[0].strength < levelManager.getStackObjectStrength(player.stack.objects[0].type)))) {
                                            disabled = true;
                                        }
                                    }

                                    if ((levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.DoubleWoodCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.DoubleMetalCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.MountedCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.SpikeCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.TNTCrate) && levelManager.survivalPoints < 40) {
                                        disabled = true;
                                    }
                                }

                                if (!disabled) {
                                    // Update the hud item.
                                    if (buildHudItem && buildHudItem.stackObjectType !== levelHudItems[i].stackObjectType) {
                                        buildData = (buildHudItem.direction === namespace.DirectionEnum.Left) ? levelManager.leftPlayer.buildData : levelManager.rightPlayer.buildData;
                                        buildData.cooldown = levelManager.getStackObjectCreationTime(levelHudItems[i].stackObjectType);
                                        buildData.stackObjectType = levelHudItems[i].stackObjectType;

                                        if (levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.DoubleWoodCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.DoubleMetalCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.MountedCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.SpikeCrate || levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.TNTCrate) {
                                            levelManager.survivalPoints -= 40;
                                        }
                                        else if (levelHudItems[i].stackObjectType === namespace.StackObjectTypeEnum.RepairCrate) {
                                            levelManager.repairKits -= 1;
                                        }
                                    }

                                    if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialHitCrateButton) {
                                        levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialBuildingFirstCrate;
                                    }

                                    clearHudSelection(true);
                                }
                                clickedHandled = true;
                            }
                            else if (currentType === namespace.CommandEnum.ShowStackObjects && playScreen) {
                                buildData = (levelHudItems[i].direction === namespace.DirectionEnum.Left) ? levelManager.leftPlayer.buildData : levelManager.rightPlayer.buildData;
                                if (levelHudItems[i].selected === false && buildData.cooldown === 0) {
                                    // Reset the selection.
                                    clearHudSelection(true);

                                    levelHudItems[i].selected = true;

                                    setHudSelection(levelHudItems[i], true);

                                    if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialHitBuildButton || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialZombiesComing) {
                                        levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialHitCrateButton;
                                    }
                                }
                                else {
                                    clearHudSelection(true);
                                }

                                clickedHandled = true;

                                namespace.SoundManager.playClickSound();
                            }
                            else if (currentType === namespace.CommandEnum.AssignWeapon) {
                                if (levelHudItems[i].locked === false) {
                                    if (levelHudItems[i].direction === namespace.DirectionEnum.Left && (levelManager.leftPlayer.weapons[levelHudItems[i].weaponCategory] !== levelHudItems[i].weaponType || levelManager.leftPlayer.currentWeaponCategory !== levelHudItems[i].weaponCategory)) {
                                        levelManager.leftPlayer.weapons[levelHudItems[i].weaponCategory] = levelHudItems[i].weaponType;
                                        //  levelManager.leftPlayer.reloadCooldown = 0.3;
                                        levelManager.leftPlayer.currentWeaponCategory = levelHudItems[i].weaponCategory;
                                    }
                                    else if (levelHudItems[i].direction === namespace.DirectionEnum.Right && (levelManager.rightPlayer.weapons[levelHudItems[i].weaponCategory] !== levelHudItems[i].weaponType || levelManager.rightPlayer.currentWeaponCategory !== levelHudItems[i].weaponCategory)) {
                                        levelManager.rightPlayer.weapons[levelHudItems[i].weaponCategory] = levelHudItems[i].weaponType;
                                        // levelManager.rightPlayer.reloadCooldown = 0.3;
                                        levelManager.rightPlayer.currentWeaponCategory = levelHudItems[i].weaponCategory;
                                    }

                                    levelManager.selectedWeaponType = levelHudItems[i].weaponType;

                                    size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });

                                    // Update the selection.
                                    for (u = 0; u < levelHudItems.length; ++u) {
                                        if (levelHudItems[u].type === namespace.CommandEnum.AssignWeapon && levelHudItems[u].direction === levelHudItems[i].direction) {
                                            if (levelHudItems[u].direction === namespace.DirectionEnum.Left) {
                                                levelHudItems[u].selected = (levelManager.leftPlayer.weapons[levelHudItems[u].weaponCategory] === levelHudItems[u].weaponType);
                                            }
                                            else {
                                                levelHudItems[u].selected = (levelManager.rightPlayer.weapons[levelHudItems[u].weaponCategory] === levelHudItems[u].weaponType);
                                            }
                                        }
                                        else if (levelHudItems[u].type === namespace.CommandEnum.Ready) {
                                            if (!levelManager.primaryWeaponAssignmentDone) {
                                                if (levelManager.leftPlayer.weapons[namespace.WeaponCategoryEnum.Primary] !== namespace.WeaponTypeEnum.None && levelManager.rightPlayer.weapons[namespace.WeaponCategoryEnum.Primary] !== namespace.WeaponTypeEnum.None) {
                                                    levelHudItems[u].ty = size.y - 1.1;
                                                }
                                                else {
                                                    levelHudItems[u].ty = size.y + 1;
                                                }
                                            }
                                            else if (!levelManager.secondaryWeaponAssignmentDone) {
                                                if (levelManager.leftPlayer.weapons[namespace.WeaponCategoryEnum.Secondary] !== namespace.WeaponTypeEnum.None && levelManager.rightPlayer.weapons[namespace.WeaponCategoryEnum.Secondary] !== namespace.WeaponTypeEnum.None) {
                                                    levelHudItems[u].ty = size.y - 1.1;
                                                }
                                                else {
                                                    levelHudItems[u].ty = size.y + 1;
                                                }
                                            }
                                        }
                                    }

                                    if (levelHudItems[i].direction === namespace.DirectionEnum.Left) {
                                        levelManager.leftPlayer.targetScaleFactor = 1.6;
                                        levelManager.rightPlayer.targetScaleFactor = 1;
                                        if (levelManager.leftPlayer.scaleFactor !== levelManager.leftPlayer.targetScaleFactor) {
                                            levelManager.leftPlayer.reloadCooldown = 0.3;
                                        }
                                    }
                                    else {
                                        levelManager.rightPlayer.targetScaleFactor = 1.6;
                                        levelManager.leftPlayer.targetScaleFactor = 1;
                                        if (levelManager.rightPlayer.scaleFactor !== levelManager.rightPlayer.targetScaleFactor) {
                                            levelManager.rightPlayer.reloadCooldown = 0.3;
                                        }
                                    }

                                    namespace.SoundManager.playReloadSound();
                                }
                                else {
                                    itemLockedDialogTime = 1.7;
                                }

                                clickedHandled = true;
                            }

                            if (clickedHandled) {
                                break;
                            }
                        }
                    }

                    if (!clickedHandled && levelManager.currentLevel === 1) {
                        if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialStart) {
                            levelManager.tutorialCooldown = 6;
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialZombiesComing;
                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstWaveDone) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialReinforceDefenses;
                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTapToContinue) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialFirstWave;
                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialSecondWaveDone) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialHelpingHand;
                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialFirstKill) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialSelectSurvivalRage;

                            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 });

                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamDefenses) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialTeamWork;
                            clickedHandled = true;

                            size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ToggleAdrenaline, "x": size.x / 2 - 0.5, "y": size.y - 2, "tx": size.x / 2 - 0.5, "ty": size.y - 2, "size": 0.95 });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 - 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 - 3, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Left });
                            levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Left });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.ShowStackObjects, "x": size.x / 2 + 1, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right });
                            levelHudItems.push({ "selected": false, "enabled": true, "type": namespace.CommandEnum.CancelAction, "x": size.x / 2 + 2, "y": size.y - 2, "size": 0.95, "direction": namespace.DirectionEnum.Right });
                            levelHudItems.push({ "selected": false, "enabled": false, "type": namespace.CommandEnum.SelectStackObject, "x": 0, "y": size.y, "tx": 0, "ty": size.y, "size": 0.95, "stackObjectType": namespace.StackObjectTypeEnum.WoodCrate, "direction": namespace.DirectionEnum.Right });
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialHelpingHand) {
                            levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialTeamCamera;

                            var position = levelManager.minX + (levelManager.maxX - levelManager.minX) / 2;
                            levelManager.leftPlayer.x = position - 0.6;
                            levelManager.leftPlayer.stack.x = position - 0.6;
                            levelManager.leftPlayer.direction = namespace.DirectionEnum.Right;
                            levelManager.leftPlayer.firingAngle = 0;
                            levelManager.leftPlayer.reloadCooldown = 0.2;
                            levelManager.leftPlayer.buildData.cooldown = 0;

                            levelManager.rightPlayer.x = position + 0.6;
                            levelManager.rightPlayer.stack.x = position + 0.6;
                            levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
                            levelManager.rightPlayer.firingAngle = 0;
                            levelManager.rightPlayer.reloadCooldown = 0.2;
                            levelManager.rightPlayer.buildData.cooldown = 0;

                            namespace.SoundManager.playReloadSound();

                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamCamera) {
                            updatePlayerDirections(point);
                            if (levelManager.leftPlayer.direction === namespace.DirectionEnum.Left && levelManager.rightPlayer.direction === namespace.DirectionEnum.Left) {
                                levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialTeamDefenses;
                            }
                            clickedHandled = true;
                        }
                        else if (levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTurnRight) {
                            var midX = (levelManager.minX + (levelManager.maxX - levelManager.minX) / 2);
                            if (point.x > midX) {
                                levelManager.leftPlayer.direction = namespace.DirectionEnum.Right;
                                levelManager.leftPlayer.firingAngle = 0;
                                levelManager.leftPlayer.reloadCooldown = 0.2;
                                levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
                                levelManager.rightPlayer.firingAngle = 0;
                                levelManager.rightPlayer.reloadCooldown = 0.2;
                                levelManager.currentLevelStage = namespace.LevelStageEnum.TutorialSecondWave;
                            }
                            clickedHandled = true;
                        }
                    }

                    if (!clickedHandled) {
                        updatePlayerDirections(point);
                    }
                    break;
                default:
            }
        }
    }

    function updatePlayerDirections(point) {
        if (currentGameState === namespace.GameStateEnum.GameRun) {
            if (!levelManager.secondaryWeaponAssignmentDone || !levelManager.primaryWeaponAssignmentDone || levelManager.newItem !== namespace.StackObjectTypeEnum.None) {
                levelManager.rightPlayer.targetScaleFactor = 1;
                levelManager.leftPlayer.targetScaleFactor = 1;
                levelManager.selectedWeaponType = namespace.WeaponTypeEnum.None;
            }
            else {
                if (levelManager.currentLevelStage === namespace.LevelStageEnum.None || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamCamera || levelManager.currentLevelStage === namespace.LevelStageEnum.TutorialTeamWork) {
                    var midX = (levelManager.minX + (levelManager.maxX - levelManager.minX) / 2);
                    if ((levelManager.leftPlayer.buildData.cooldown > 0 || levelManager.rightPlayer.buildData.cooldown > 0) && !(levelManager.leftPlayer.buildData.cooldown > 0 && levelManager.rightPlayer.buildData.cooldown > 0)) {
                        if (point.x < midX) {
                            if (levelManager.leftPlayer.direction !== namespace.DirectionEnum.Left) {
                                levelManager.leftPlayer.direction = namespace.DirectionEnum.Left;
                                levelManager.leftPlayer.firingAngle = 0;
                                levelManager.leftPlayer.reloadCooldown = 0.2;
                            }

                            if (levelManager.rightPlayer.direction !== namespace.DirectionEnum.Left) {
                                levelManager.rightPlayer.direction = namespace.DirectionEnum.Left;
                                levelManager.rightPlayer.firingAngle = 0;
                                levelManager.rightPlayer.reloadCooldown = 0.2;
                            }
                        }
                        else {
                            if (levelManager.leftPlayer.direction !== namespace.DirectionEnum.Right) {
                                levelManager.leftPlayer.direction = namespace.DirectionEnum.Right;
                                levelManager.leftPlayer.firingAngle = 0;
                                levelManager.leftPlayer.reloadCooldown = 0.2;
                            }

                            if (levelManager.rightPlayer.direction !== namespace.DirectionEnum.Right) {
                                levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
                                levelManager.rightPlayer.firingAngle = 0;
                                levelManager.rightPlayer.reloadCooldown = 0.2;
                            }
                        }
                    }
                    else if (point.x < midX && levelManager.leftPlayer.direction !== namespace.DirectionEnum.Left) {
                        levelManager.leftPlayer.direction = namespace.DirectionEnum.Left;
                        levelManager.leftPlayer.firingAngle = 0;
                        levelManager.leftPlayer.reloadCooldown = 0.2;
                    }
                    else if (point.x < midX && levelManager.rightPlayer.direction !== namespace.DirectionEnum.Left) {
                        levelManager.rightPlayer.direction = namespace.DirectionEnum.Left;
                        levelManager.rightPlayer.firingAngle = 0;
                        levelManager.rightPlayer.reloadCooldown = 0.2;
                    }
                    else if (point.x > midX && levelManager.rightPlayer.direction !== namespace.DirectionEnum.Right) {
                        levelManager.rightPlayer.direction = namespace.DirectionEnum.Right;
                        levelManager.rightPlayer.firingAngle = 0;
                        levelManager.rightPlayer.reloadCooldown = 0.2;
                    }
                    else if (point.x > midX && levelManager.leftPlayer.direction !== namespace.DirectionEnum.Right) {
                        levelManager.leftPlayer.direction = namespace.DirectionEnum.Right;
                        levelManager.leftPlayer.firingAngle = 0;
                        levelManager.leftPlayer.reloadCooldown = 0.2;
                    }
                }

                // Reset the hud items.
                // clearHudSelection(true);
            }
        }
    }

    function setHudSelection(mainHudItem, animate) {
        var offsetThrowableItems = mainHudItem.x, offsetCrates = mainHudItem.x, isCrate = false;
        for (var u = 0; u < levelHudItems.length; ++u) {
            if (levelHudItems[u].type === namespace.CommandEnum.SelectStackObject && levelHudItems[u].direction === mainHudItem.direction) {
                if (!animate) {
                    levelHudItems[u].y = mainHudItem.y + 1;
                }
                levelHudItems[u].ty = mainHudItem.y + 1;
                levelHudItems[u].enabled = true;

                switch (levelHudItems[u].stackObjectType) {
                    case namespace.StackObjectTypeEnum.WoodCrate:
                    case namespace.StackObjectTypeEnum.MetalCrate:
                    case namespace.StackObjectTypeEnum.RepairCrate:
                        if (levelHudItems[u].direction === namespace.DirectionEnum.Left) {
                            offsetCrates -= 1;
                        }
                        else {
                            offsetCrates += 1;
                        }
                        isCrate = true;
                        break;
                    case namespace.StackObjectTypeEnum.DoubleWoodCrate:
                    case namespace.StackObjectTypeEnum.DoubleMetalCrate:
                    case namespace.StackObjectTypeEnum.MountedCrate:
                    case namespace.StackObjectTypeEnum.TNTCrate:
                    case namespace.StackObjectTypeEnum.SpikeCrate:
                        if (levelHudItems[u].direction === namespace.DirectionEnum.Left) {
                            offsetThrowableItems += 1;
                        }
                        else {
                            offsetThrowableItems -= 1;
                        }
                        isCrate = false;
                        break;
                }

                if (levelHudItems[u].direction === namespace.DirectionEnum.Left) {
                    levelHudItems[u].x = (isCrate ? offsetCrates + 1 : offsetThrowableItems);
                    levelHudItems[u].tx = levelHudItems[u].x;
                }
                else {
                    levelHudItems[u].x = (isCrate ? offsetCrates - 1 : offsetThrowableItems);
                    levelHudItems[u].tx = levelHudItems[u].x;
                }
            }
        }
    }

    function clearHudSelection(animate) {
        var size = namespace.screenToHud({ "x": canvas.width, "y": canvas.height });
        var i, len, hudItem;
        for (i = 0, len = levelHudItems.length; i < len; i += 1) {
            hudItem = levelHudItems[i];
            if (hudItem.type === namespace.CommandEnum.ShowStackObjects) {
                hudItem.selected = false;
            }
            else if (hudItem.type === namespace.CommandEnum.SelectStackObject) {
                hudItem.enabled = false;
                if (!animate) {
                    hudItem.y = size.y;
                }
                hudItem.ty = size.y;
            }
        }
    }

    function pointerUp() {
        isPointerDown = false;
    }

})(window.ZCJGame = window.ZCJGame || {});
